from sage.rings.rational_field import QQ
from sage.matrix.constructor import matrix, vector
from sage.matrix.special import block_matrix

import admcycles.admcycles
from .admcycles import list_strata, generating_indices, pullback_matrix


def identify_class(tautring, d, pairfunction=None, pullfunction=None,
                   kappapsifunction=None, check=False):
    r"""
    Identify a tautological class in degree d by specifying its
    intersection pairing with classes in complementary degree
    and/or its pullback under boundary gluing morphisms.

    INPUT::

        - tautring (TautologicalRing) : ring containing the class that
          is to be identified

        - d (integer): degree of the class

        - pairfunction (function, optional): function taking classes in
          complementary degree and returning the intersection product with
          the desired class

        - pullfunction (function, optional): function taking stable graphs
          with at most two vertices and returning the prodtautclass obtained
          by pulling back the desired class under the associated boundary
          gluing map

        - kappapsifunction (function, optional): function taking as arguments
          a list, dictionary and TautologicalClass (giving a polynomial in
          kappa and psi-classes) and returning the intersection product of
          the desired class with this polynomial (or None)

        - check (bool, default = False) : whether to use FZ relations to
          verify that intersection numbers/pullbacks are consistent with
          known tautological relations

    EXAMPLES::

        sage: from admcycles.identify_classes import identify_class
        sage: from admcycles import TautologicalRing
        sage: R = TautologicalRing(2, 1)
        sage: cl = 3 * R.psi(1) -2 * R.kappa(1)
        sage: def pairf(a):
        ....:     return (a * cl).evaluate()
        sage: cl2 = R.identify_class(1, pairfunction = pairf)
        sage: cl == cl2
        True
        sage: def pullf(gamma):
        ....:     return gamma.boundary_pullback(cl)
        ....:
        sage: cl3 = R.identify_class(1, pullfunction=pullf)
        sage: cl == cl3
        True

    Here is some code to reconstruct the class 42*psi_1 on Mbar_0,6 from its intersection
    numbers with psi-monomials (which have an explicit formula). Note that the associated
    function kpfun is allowed to return None if kappa-classes show up::

        sage: R = TautologicalRing(0,6)
        sage: def kpfun(kap, psi, cl):
        ....:     if kap:
        ....:         return None
        ....:     b = vector(ZZ,[psi.get(i,0) for i in range(1,7)])+vector([1,0,0,0,0,0])
        ....:     return 42*factorial(3)/prod(factorial(bi) for bi in b)
        ....:
        sage: cl4 = R.identify_class(1, kappapsifunction=kpfun)
        sage: cl4 == 42*R.psi(1)
        True

    TESTS::

        sage: R = TautologicalRing(2, 1)
        sage: def bad_pairf(a):
        ....:     return (a * cl).evaluate() if not (a - R.kappa(3)).is_empty() else 0
        sage: R.identify_class(1, pairfunction = bad_pairf, check=True)
        Traceback (most recent call last):
        ...
        ValueError: intersection numbers or pullbacks are not consistent with FZ relations
    """
    gens = tautring.basis(d)
    ngens = len(gens)
    d_dual = tautring.socle_degree() - d
    M = matrix(QQ, 0, ngens)
    v = []
    rank_sufficient = (ngens == 0)
    if pairfunction is not None and not rank_sufficient:
        if check:
            gens_dual = tautring.generators(d_dual)
        else:
            gens_dual = tautring.basis(d_dual)
        ind_dcomp = tuple(generating_indices(tautring._g, tautring._n, d))
        inters = tautring.pairing_matrix(d_dual, basis=not check, ind_dcomp=ind_dcomp)
        M = block_matrix([[M], [inters]], subdivide=False)
        v += [pairfunction(a) for a in gens_dual]
        rank_sufficient = (M.rank() == ngens)

    if pullfunction is not None and not rank_sufficient:
        moduli = tautring._moduli
        graphs = [gam for gam in list_strata(tautring._g, tautring._n, 1)
                  if not gam.vanishes(moduli)]
        if len(graphs) > 0 and graphs[0].num_verts() == 1:
            irrgraph = graphs.pop(0)
            graphs.append(irrgraph)
        for gam in graphs:
            pulls = pullback_matrix(tautring._g, tautring._n, d, bdry=gam)
            # pulls = [gam.boundary_pullback(a).totensorTautbasis(d, vecout=True)
            #          for a in gens]
            M = block_matrix([[M], [pulls]],
                             subdivide=False)
            v += list(pullfunction(gam).totensorTautbasis(d, vecout=True))
            rank_sufficient = (M.rank() == ngens)
            if rank_sufficient and not check:
                break

    if kappapsifunction is not None and not rank_sufficient:
        for kappa, psi, cl in tautring.kappa_psi_polynomials(d_dual, True):
            intnum = kappapsifunction(kappa, psi, cl)
            if intnum is not None:
                M = block_matrix(QQ, [[M], [matrix([[(c * cl).evaluate() for c in gens]])]], subdivide=False)
                v.append(intnum)
            rank_sufficient = (M.rank() == ngens)
            if rank_sufficient and not check:
                break

    if not rank_sufficient:
        raise ValueError('Intersection numbers and/or pullbacks are ' +
                         'not enough to determine this class')
    try:
        w = M.solve_right(vector(QQ, v))
    except ValueError:
        raise ValueError('intersection numbers or pullbacks are not consistent with FZ relations')
    return tautring.from_basis_vector(w, d)


def Pullback_Matrices(g, n, d, ind_list=None):
    r"""
    Compute the pullback matrices of given degree d and list of
    boundary stratum.

    INPUT::


        - ind_list (list): the list of indices of the boundary stratum,
                           0 indicates the first nontrivial graph
                           appearing in the generator
                           list of degree d tautological classes

    EXAMPLES::

    sage: from admcycles.identify_classes import Pullback_Matrices
    sage: from admcycles import TautologicalRing
    sage: Pullback_Matrices(1, 2, 1)
    [
    [1]  [1]
    [0], [1]
    ]

    """
    L1 = admcycles.admcycles.tautgens(g, n, d)
    Listgen1 = admcycles.admcycles.generating_indices(g, n, d)
    L2 = admcycles.admcycles.tautgens(g, n, 1)

    M_list = []
    if not ind_list:
        ind_list = range(len(L2) - n - 1)

    for a in ind_list:
        b = a + n + 1
        M0 = []
        Q = None
        if b == len(L2) - 1:
            Q = list(L2[b].standard_markings()._terms)[0]
        else:
            Q = list(L2[b]._terms)[0]

        for i in Listgen1:
            T = Q.boundary_pullback(L1[i])
            M0 += [T.totensorTautbasis(d, vecout=True)]
        M = matrix(M0)
        M_list += [M]
    return M_list


def solve_perfect_pairing(g, n, d, list_evaluate):
    r"""
    Identify a tautological class in degree d by specifying
    its intersection pairing
    with classes in complementary degree.

    INPUT::


        - list_evaluate (list) : the list of intersection numbers

        - list_dual_classes (list): a list of classes of dual codimension
                                    one wants to compute the intersection
                                    (the program will complete the list
                                    such that one obtain a full rank matrix)


    EXAMPLES::

        sage: from admcycles.identify_classes import solve_perfect_pairing
        sage: from admcycles import TautologicalRing
        sage: R = TautologicalRing(1,3)
        sage: cl = R.psi(1) -2 * R.kappa(1)
        sage: def pairf(a):
        ....:     return (a * cl).evaluate()
        sage: cl2 = solve_perfect_pairing(1, 3, 1,
        ....: [pairf(x) for x in R.generators(2)])
        sage: cl == cl2
        True

    """
    d_dual = 3 * g - 3 + n - d
    ind = admcycles.admcycles.generating_indices(g, n, d)
    gen = admcycles.admcycles.tautgens(g, n, d)
    gen_dual = admcycles.admcycles.tautgens(g, n, d_dual)
    M_intpair = []

    for i in ind:
        intersection_vector = [(gen[i] * X).evaluate() for X in gen_dual]
        M_intpair.append(intersection_vector)
    Sol = list(matrix(M_intpair).solve_left(matrix(list_evaluate)))[0]
    output = sum(Sol[i] * gen[ind[i]] for i in range(0, len(ind)))
    return output


def solve_clutch_pullback(g, n, d, bdiv_indices, pullback_classes):
    r"""
    Identify a tautological class in degree d by specifying its pullback
    under boundary gluing morphisms.

    INPUT::


        - bdiv_indices(list) : the list of indices of boundary
                               stratum(the first nontrivial graph in
                               tautgen list will be indexed
                               by 0) to which we will pullback
        - pullback_classes(list) : the list of pullbacks of the tautological
                                    class with respect to the bdiv listed

    EXAMPLES::

        sage: from admcycles.identify_classes import solve_clutch_pullback
        sage: from admcycles import TautologicalRing
        sage: R = TautologicalRing(1,3)
        sage: stgraphs = [list(x._terms)[0] for x in R.generators(1)[4:]]
        sage: cl = R.psi(1) -2 * R.kappa(1)
        sage: def pullf(gr):
        ....:     return gr.boundary_pullback(cl)
        sage: cl2 = solve_clutch_pullback(1, 3, 1, range(len(stgraphs)),
        ....: [pullf(gr) for gr in stgraphs])
        sage: cl == cl2
        True

    """

    if 2 * d > 2 * g - 2 + n:
        print(" It may not work because the codim is too large")
    ListV = [T.totensorTautbasis(d, vecout=True) for T in pullback_classes]
    ListM = Pullback_Matrices(g, n, d)
    ListM_1 = [ListM[i] for i in bdiv_indices]
    Sol = SolveMultLin(ListM_1, ListV)
    ind = admcycles.admcycles.generating_indices(g, n, d)
    gen = admcycles.admcycles.tautgens(g, n, d)
    output = sum(Sol[i] * gen[ind[i]] for i in range(len(ind)))
    return output


def SolveMultLin(ListM, ListV):

    dim = ListM[0].nrows()
    assert all(Q.nrows() == dim for Q in ListM)
    NewM = []
    NewV = []
    for Q in ListM:
        NewM += list(Q.transpose())
    NewM = matrix(NewM).transpose()
    for W in ListV:
        NewV += list(W)
    NewV = matrix([NewV])
    result = NewM.solve_left(NewV)
    L = list(result)
    Sol = vector(L[0])
    return Sol



from sage.all import cached_function, QQ, PermutationGroup, Partitions, prod
from admcycles import psiclass, lambdaclass, list_strata, Hyperell, TautologicalRing, Strataclass, Wittenrspin
from admcycles.admcycles import common_degenerations
from admcycles.diffstrata import Stratum
from admcycles.double_ramification_cycle import DR_cycle
import admcycles.diffstrata.cache

# ASV forks before running a benchmark so all caches are empty when a benchmark
# function is first run. However, asv does not fork between repetitions of a
# benchmark so make sure to always run asv with --quick. We could fix this by
# decorating each benchmark or suite such that it forks again; wouldn't be hard
# if we believe that this makes sense.
@cached_function
def nothing_is_cached():
    return None


def to_exp_dict(p):
    d = {}
    for i in p:
        if i in d:
            d[i] += 1
        else:
            d[i] = 1
    return d


def pixton_list_strata(g, n, r, moduli='st'):
    r"""
    Enumerate all stable graphs of given genus g, number n of marked points and
    moduli.

    This is currently done via the subpackage DR which supports the moduli
    option.
    """
    from admcycles.DR.graph import all_pure_strata
    from admcycles.admcycles import Graphtodecstratum
    from admcycles.moduli import get_moduli
    MODU = get_moduli(moduli, DRpy=True)
    markings = tuple(range(1, n + 1))
    L = all_pure_strata(g, r, markings, MODU)
    return [Graphtodecstratum(t).gamma for t in L]


class NoProduct:
    r"""
    When a list of .params starts with this entry, not all combinations of the
    entries are processed but each entry is understood as a list of parameters.
    """
    def __init__(self, names):
        self._names = names

    def __repr__(self):
        return self._names


class Benchmark:
    r"""
    Base class for benchmarks that understand the NoProduct marker defined above.
    """
    def setup(self, *args):
        if args and isinstance(args[0], NoProduct):
            raise NotImplementedError("this marker input will be ignored by ASV because we are throwing a NotImplementedError in setup()")


class StrataGeneration(Benchmark):
    def time_list_strata(self, gn):
        g, n = gn
        list_strata(g, n, 3*g-3+n)

    time_list_strata.params = [NoProduct('g, n'), (0, 7), (1, 5), (2, 3), (3, 2), (4, 0)] # (2, 4) times out


class CommonDegenerations(Benchmark):
    def setup(self, *args):
        super().setup(*args)
        g, n, mod, m1, m2 = args[0]
        self.L1 = pixton_list_strata(g, n, m1, mod)
        self.L2 = pixton_list_strata(g, n, m2, mod)

    def time_common_degenerations(self, gnmodm1m2):
        for gam1 in self.L1:
            for gam2 in self.L2:
                common_degenerations(gam1, gam2, modiso=True)

    time_common_degenerations.params = [NoProduct('g, n, mod, m1, m2'), (0, 6, 'st', 2, 2), (0, 6, 'st', 2, 3), (2, 3, 'st', 3, 3), (3, 2, 'ct', 4, 2)]


class FZRelations(Benchmark):
    def time_basis(self, gnrmod):
        g, n, r, mod = gnrmod
        R = TautologicalRing(g, n, moduli=mod)
        L = R.basis(r)

    time_basis.params = [NoProduct('g, n, r, mod'), (4, 1, 4, 'st'), (5, 0, 5, 'ct'), (3, 4, 3, 'rt'), (12, 2, 8, 'sm')]


class Evaluation(Benchmark):
    def time_psi_classes(self, gn):
        g, n = gn
        for p in Partitions(3*g-3+n, max_length=n):
            prod(psiclass(i+1, g, n)**j for i,j in enumerate(p)).evaluate()

    time_psi_classes.params = [NoProduct('g, n'), (0, 33), (1, 18), (2, 16), (6, 10), (9, 7), (13, 5), (20, 4)]


class AlgebraInTautologicalRing(Benchmark):
    def time_lambda_and_psi_product1(self):
        # from https://gitlab.com/modulispaces/admcycles/issues/28
        lambdaclass(3,3,3) * psiclass(1,3,3)**2 * psiclass(2,3,3)**2 * psiclass(3,3,3)**2

    def time_lambda_power(self, dgnp):
        d, g, n, p = dgnp
        lambdaclass(d,g,n)**p

    time_lambda_power.params = [NoProduct('d, g, n, p'), (1, 3, 2, 3), (2, 4, 0, 2)]


class PairingMatrix(Benchmark):
    def setup(self, *args):
        super().setup(*args)

        g, n, mod, m1, m2 = args[0]
        R = TautologicalRing(g, n, moduli=mod)
        self.gens = R.generators(m1)
        self.cogens = R.generators(m2)

    def time_pairing_matrix(self, gnmodm1m2):
        [[a * b for a in self.gens] for b in self.cogens]

    time_pairing_matrix.params = [NoProduct('g, n, mod, m1, m2'), (3, 0, 'st', 3, 2), (2, 2, 'ct', 3, 2)]


class TautologicalClasses(Benchmark):
    def time_lambda(self, dgn):
        d, g, n = dgn
        lambdaclass(d,g,n)

    time_lambda.params = [NoProduct('d, g, n'), (2, 2, 8), (3, 4, 3), (4, 4, 0)]

    def time_double_ramification_cycle(self, gAk):
        g, A, k = gAk
        DR_cycle(g, A, k=k)

    time_double_ramification_cycle.params = [NoProduct('g, A, k'),
                                             (2, (3, 1, 1, 1, 1, 1), 1),
                                             (3, (4, 4, -1), 1),
                                             (3, (5, 1, 1, 1), 1)]

    def time_hyperelliptic_locus(self, gnm):
        g, n, m = gnm
        Hyperell(g, n, m)

    time_hyperelliptic_locus.params = [NoProduct('g, n, m'), (2, 1, 0), (3, 0, 0), (2, 0, 1)]

    def time_strataclass(self, gkmuargs):
        Strataclass(*gkmuargs)

    time_strataclass.params = [NoProduct('g, k, mu, **args'),
                               (3, 2, (3, 5)),
                               (2, 1, (2, 4, -2, -2), False, (2,), 0, 'diffstrata'),
                               (2, 1, (8, -2, -4), False, (), 0, 'diffstrata', True, False)]

    def time_witten_rspin(self, gAargs):
        Wittenrspin(*gAargs)

    time_witten_rspin.params = [NoProduct('g, A, **args'),
                                (3, (1,1,1,1)),
                                (2, (5, (QQ((1, 2)), -1), (QQ((1, 2)), -2))),
                                (3, (5,6), 7)]


class AbelianDifferentials(Benchmark):
    def time_euler_characteristic(self, sig):
        admcycles.diffstrata.cache.ADM_EVALS = admcycles.diffstrata.cache.FakeCache()
        admcycles.diffstrata.cache.TOP_XIS = admcycles.diffstrata.cache.FakeCache()
        X = Stratum(sig)
        X.euler_characteristic()

    time_euler_characteristic.params = [NoProduct('Signature'), (3, 1), (2, -3, 3), (50, -52, 2)]

    def time_top_xi_powers(self, sig):
        admcycles.diffstrata.cache.ADM_EVALS = admcycles.diffstrata.cache.FakeCache()
        admcycles.diffstrata.cache.TOP_XIS = admcycles.diffstrata.cache.FakeCache()
        X = Stratum(sig)
        X.top_xi_at_level(((), 0), 0)

    time_top_xi_powers.params = [NoProduct('Signature'), (8,), (1, -3, 4), (2, -3, 3), (80, -82, 1, 1)]



r"""
Spin Stratum: A stratum with signature of even type will has spin structure, i.e.
its connected components can be partitioned to two non-empty sets.

cf. [KZ03]_ and [Boi15]_

"""

from copy import deepcopy
import sage.misc.persist

from sage.rings.all import QQ
from sage.matrix.constructor import matrix, vector
from sage.misc.cachefunc import cached_function
from sage.combinat.all import SetPartitions

import admcycles.admcycles
import admcycles.stratarecursion

import admcycles.diffstrata.levelgraph
import admcycles.diffstrata.bic
import admcycles.diffstrata.stratatautring
import admcycles.diffstrata.embeddedlevelgraph

from admcycles.stable_graph import StableGraph
from admcycles.diffstrata.generalisedstratum import GeneralisedStratum
from admcycles.diffstrata.additivegenerator import AdditiveGenerator
from admcycles.diffstrata.elgtautclass import ELGTautClass
from admcycles.diffstrata.sig import Signature

from admcycles.identify_classes import SolveMultLin


def Spin_strataclass(sig, res_cond=[]):
    r"""
    Compute the spin stratum class of given signature of even type (may
    have paired simple poles) and residue conditions.

    INPUT:

    - sig (tuple)
    - res_cond (list): It is a list of lists corresponding to residue
      conditions. For example, sig=(4,-2,-1,-1) with the
      simple poles being paired. Then
      res_cond=[[(0,2), (0,3)]]. Here the zero indicate
      the components of a generalized stratum.

    EXAMPLES::

        sage: from admcycles.diffstrata.spinstratum import Spin_strataclass
        sage: Spin_strataclass((2,2)).basis_vector()
        (159/4, -179/48, -7/24, -7/24, -131/48, -23/24, -131/48, 149/24, -83/16, 271/48, 77/48, 77/48, -193/8, -185/48, 127/48, 395/48, 221/48, -73/8, -185/48, 127/48, 395/48, 221/48, -73/8, -185/48, -41/48, 389/48, 389/48, -23/8, 51/8, -139/16, -323/16, 1/8, -25/4, -11/4, -11/4, -23/4, 973/48, 37/96, -5/2, 23/96, 23/96, -25/32)


    """
    X = SpinStratum([Signature(sig)], res_cond=res_cond)
    A = AG_with_spin(X, ((), 0))
    E = ELGT_with_spin(X, [(1, A)])
    return E.to_prodtautclass_spin().pushforward()


#######################################################################
#######################################################################
# Define the class of spin strata #####################################
#######################################################################

class SpinStratum(GeneralisedStratum):

    def __init__(self, sig_list, res_cond=None, pair_res_cond=None):
        r"""
        A stratum can have spin structure if there are pairs of simple poles,
        such that their residues r_i + r_i+1 = 0. In this code, we only handle the
        cases of at most 1 pair of simple poles.

        INPUT:

        - sig_list (list): a list of tuples indicating the signatures
          of the components of a generalised stratum

        - res_cond (list): a list whose entries are lists corresponding
          to residue conditions

        EXAMPLES::

            sage: from admcycles.diffstrata.spinstratum import SpinStratum
            sage: from admcycles.diffstrata.sig import Signature
            sage: X = SpinStratum([Signature((4,-2,-1,-1))],
            ....: res_cond=[[(0,2),(0,3)]])

        """

        super().__init__(sig_list, res_cond)

        if res_cond:
            self._res_cond = res_cond
        else:
            self._res_cond = []
        if pair_res_cond:
            self.pair_res_cond = pair_res_cond
        else:
            self.pair_res_cond = []
        self.odd_sing_list = []
        for i in range(0, len(self._sig_list)):
            for j in range(0, len(sig_list[i].sig)):
                if sig_list[i].sig[j] % 2 != 0:
                    self.odd_sing_list.append((i, j))

        self.sim_pole_list = list(self.simple_poles())
        self.sim_pole_list.sort()
        self.hor_edge = []
        self.eff_hor_edge = []
        self.eff_pair_res = []

        if self.odd_sing_list != [] and pair_res_cond is None:
            self.is_legal()

    def __repr__(self):
        return "SpinStratum(sig_list=%r,res_cond=%r)" % (
            self._sig_list, self._res_cond)

    def __str__(self):
        rep = ''
        if self._h0 > 1:
            rep += 'Product of Strata with spin:\n'
        else:
            rep += 'Spin Stratum: '
        for sig in self._sig_list:
            rep += repr(sig.sig) + '\n'
        rep += 'with residue conditions: '
        if not self._res_cond:
            rep += repr([]) + '\n'
        for res in self._res_cond:
            rep += repr(res) + '\n'
        return rep

    def is_legal(self):
        r"""
        This is just a method for initializing a spin stratum, checking whether
        the stratum really has spin structure.

        EXAMPLES::

            sage: from admcycles.diffstrata.spinstratum import SpinStratum
            sage: from admcycles.diffstrata.sig import Signature
            sage: X = SpinStratum([Signature((4,-2,-2))])
            sage: X = SpinStratum([Signature((4,-3,-1))])
            Traceback (most recent call last):
            ...
            ValueError: The Stratum does not have spin components
            sage: X = SpinStratum([Signature((4,-2,-1,-1))])
            sage: X = SpinStratum([Signature((4,-1,-1,-1,-1))])
            Traceback (most recent call last):
            ...
            NotImplementedError: We can only handle one pair of simple poles

        """

        if len(self.odd_sing_list) - len(self.sim_pole_list) > 0:
            raise ValueError("The Stratum does not have spin components")
        elif len(self.sim_pole_list) > 0:
            t = len(self.sim_pole_list)
            if t > 2:
                raise NotImplementedError("We can only handle " +
                                          "one pair of simple poles")
            if not t % 2 == 0:
                raise ValueError("The Stratum does not have spin components")

            M1 = self.smooth_LG.residue_matrix_from_RT
            if self.res_cond != []:
                M2 = M1.stack(self.matrix_from_res_conditions(self.res_cond))
            else:
                M2 = M1
            a = 0
            b = M2.rank()
            c = M1.rank()

            for q in range(0, t - 1):
                for q1 in range(q + 1, t):
                    if self.sim_pole_list[q][0] != self.sim_pole_list[q1][0]:
                        res = [self.sim_pole_list[q], self.sim_pole_list[q1]]
                        T = self.matrix_from_res_conditions(res)
                        M0 = M2.stack(T)
                        a = M0.rank()
                        if a > b:
                            continue
                        else:
                            self.hor_edge.append([self.sim_pole_list[q],
                                                  self.sim_pole_list[q1]])
            sim_pole_list_partitons = SetPartitions(t, [2 for i in
                                                    range(0, int(t / 2))]
                                                    ).list()
            rc = []
            for part in sim_pole_list_partitons:
                pair_res_cond_temp = []
                for q in part:
                    w = list(q)
                    pair_res_cond_temp.append([self.sim_pole_list[i - 1]
                                               for i in w])

                M0 = M2.stack(self.matrix_from_res_conditions(
                              pair_res_cond_temp))
                a = M0.rank()
                if a > b:
                    continue
                else:
                    self.pair_res_cond = pair_res_cond_temp
                    rc = pair_res_cond_temp + deepcopy(self._res_cond)
                    break

            if self.pair_res_cond is None:
                raise ValueError("The Stratum does not have spin components")
            if t == 1 and a == 1:
                self._res_cond = []
            elif not self.is_empty():
                new_rc = []
                for j, q in enumerate(rc):
                    rk = M1.stack(self.matrix_from_res_conditions([q])).rank()
                    if rk > c:
                        M1 = M1.stack(self.matrix_from_res_conditions([q]))
                        c += 1
                        if q in self.pair_res_cond:
                            self.eff_pair_res.append(j)
                        new_rc.append(q)
                new_rc.reverse()
                self._res_cond = new_rc
        return

    def remove_pair_res(self, psis=None):
        r"""
        This method will return the psi polynomial of the stratum in the
        tautological ring of ambient stratum (without a residue condition
        on the pair of simple poles. The formula follows Sauvaget-19 and we pick
        the leg of a simple pole for the elimination of psi class.

        Only the BICs, which have some odd enhancement (so odd even are half
        half), or still preserve the residue condition of the pair of
        simple poles, are left. Hence, one can
        still recursively compute the spin classes.

        INPUT:

        - psis (list): a list of dicts indicating the psi polynomial

        EXAMPLES::


            sage: from admcycles.diffstrata.spinstratum import SpinStratum
            sage: from admcycles.diffstrata.sig import Signature
            sage: from admcycles.admcycles import psiclass
            sage: X = SpinStratum([Signature((4,-2,-1,-1))],res_cond=[[(0,2),(0,3)]])
            sage: cl1 = X.remove_pair_res().to_prodtautclass_spin().pushforward()
            sage: (psiclass(1,1,4)**2*cl1).evaluate()
            5/24

        """
        if not self.eff_pair_res == []:
            # print("removing pair res")
            res_cond = deepcopy(self._res_cond)
            lost_rc = res_cond.pop()
            leg = lost_rc[0]
            new_stratum = SpinStratum(self._sig_list, res_cond,
                                      pair_res_cond=self.pair_res_cond)
            new_AG = new_stratum.additive_generator(((), 0), psis)
            elgt = new_AG.as_taut() * (new_stratum.res_stratum_class(lost_rc) +
                                       new_stratum.xi_with_leg() -
                                       new_stratum.xi_with_leg(leg))
            temp_psi_list = []
            for a in elgt._psi_list:
                if all(kappa % 2 != 0
                       for e, kappa in a[1]._G.LG.prongs_list()):
                    temp_psi_list.append(a)
            psi_list = [(c, AG_with_spin(new_stratum,
                        AG._enh_profile, AG._leg_dict))
                        for c, AG in temp_psi_list]
            elgt_odd_enh = ELGT_with_spin(new_stratum, psi_list)
            # print(str(elgt_odd_enh))
            result = elgt_odd_enh
            return result
        else:
            raise NotImplementedError

    def smooth_con_stratum_prodtautclass(self):
        r"""
        To compute the spin stratum class of a connected stratum. The output will be a
        product tautological class. By making use of standised signature, we will do
        the core computation in the function smooth_stratum_standard().

        EXAMPLES::

            sage: from admcycles.diffstrata.spinstratum import SpinStratum
            sage: from admcycles.diffstrata.sig import Signature
            sage: X=SpinStratum([Signature((4,0,-2))])
            sage: X.smooth_con_stratum_prodtautclass().pushforward().basis_vector()
            (-63/2, 21/2, -20, -63/2, -37/2, 38, 61/2, 61/2, 147/2, 29, 57/2, 29, -41, -60, 21, -63/2, -42, 61/2, -41, -139/2, 147/2, 63/2, -147/2, 21/2, 0, -21/2, -7, -21/2, -19/2, 19/2, 21/2, -21/2, 1, -21/2, 0, 0, 21/2, 21/2, -7, 1, -1, -1, 0, 7)

        """
        res_cond = self._res_cond
        sig = self._sig_list[0].sig
        g = self._sig_list[0].g
        LG = self.smooth_LG.LG
        stgraph0 = LG.stgraph

        if all(x == 0 for x in sig):  # it is just g=1 and the stratum class is just fundclass of M_1,n
            t = deepcopy(smooth_stratum_standard(sig))
            if all(w == 0 for w in t[3]):
                result = admcycles.admcycles.prodtautclass(stgraph0, terms=[])
            else:
                ind = admcycles.admcycles.generating_indices(t[0], t[1], t[2])
                gen = admcycles.admcycles.tautgens(t[0], t[1], t[2])
                result = sum([t[3][i] * gen[ind[i]]
                              for i in range(0, len(t[3]))]).toprodtautclass()
            return result

        L = standardise_sig(sig, res_cond)  # We use the standardised signature
        simplified = L[0]
        num_of_reg = L[1]
        legs = {x2: x1 for x1, x2 in L[2].items()}
        new_rc = L[3]

        # The stratum classes of signature with 0s are essentially just the
        # forgetful pullback of stratum class of signature without 0s.
        # We use the function smooth_stratum_standard to do the pullbacks solving.

        t = deepcopy(smooth_stratum_standard(simplified, tuple(new_rc)))
        if all(w == 0 for w in t[3]):
            cl = admcycles.admcycles.fundclass(g, len(simplified))
        else:
            ind = admcycles.admcycles.generating_indices(t[0], t[1], t[2])
            gen = admcycles.admcycles.tautgens(t[0], t[1], t[2])
            cl = sum([t[3][i] * gen[ind[i]] for i in range(0, len(t[3]))])
        if num_of_reg != 0:
            cl = cl.forgetful_pullback([len(simplified) + i
                                        for i in range(1, L[1] + 1)])
        result = cl.rename_legs(legs, inplace=False).toprodtautclass()
        # print(str(legs)+ "\n"+ str(cl) + "\n" + str(result))
        return result

#######################################################################
#######################################################################
# Define the spin version additive generators #########################
#########################################
# * the spin taut class of those enhanced level graphs with some even
#   enhancement are just considered as zero in this algorithm, because
#   the pushforward of them to \bar M_g,n will be zero.
#######################################################################


class AG_with_spin (AdditiveGenerator):

    def __init__(self, X, enh_profile, leg_dict=None):
        assert isinstance(X, SpinStratum)
        super().__init__(X, enh_profile, leg_dict)
        self.X = X

    def __repr__(self):
        return "AG_with_spin(X=%r,enh_profile=%r,leg_dict=%r)"\
            % (self._X, self._enh_profile, self._leg_dict)

    def __str__(self):
        str = ""
        if self._leg_dict is not None:
            for l in self._leg_dict:
                str += "Spin Psi class %r with exponent %r on level %r * "\
                    % (l, self._leg_dict[l], self._level_dict[l])
        str += "Graph %r" % (self._enh_profile,)
        return str

    def to_prodtautclass_spin(self, rearrange_markings=True):
        r"""
        This function is to compute the product tautological class of an
        additive generator with spin. It is basically the same as
        toprodtautclass, but just includes more case divisions.

        EXAMPLES::

            sage: from admcycles.diffstrata.spinstratum import SpinStratum,AG_with_spin
            sage: from admcycles.diffstrata.sig import Signature
            sage: X = SpinStratum([Signature((4,-2))])
            sage: AG_with_spin(X, ((1,),0)).to_prodtautclass_spin()
            Outer graph : [1, 0] [[3, 4], [1, 2, 5, 6]] [(3, 5), (4, 6)]
            Vertex 0 :
            Graph :      [1] [[1, 2]] []
            Polynomial : -1/2
            Vertex 1 :
            Graph :      [0] [[1, 2, 3, 4]] []
            Polynomial : psi_4
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1] [[1, 2]] []
            Polynomial : -1/2
            Vertex 1 :
            Graph :      [0, 0] [[2, 3, 11], [1, 4, 12]] [(11, 12)]
            Polynomial : 3
            <BLANKLINE>
            <BLANKLINE>
            Vertex 0 :
            Graph :      [1] [[1, 2]] []
            Polynomial : -1/2
            Vertex 1 :
            Graph :      [0, 0] [[3, 4, 11], [1, 2, 12]] [(11, 12)]
            Polynomial : -3

        """
        LG = self._G.LG
        stgraph = LG.stgraph

        # If the level graph has any edge with positive even enhancement,
        # half of the prong matchings give an even spin curve and half of
        # them give an odd spin curve
        if any(kappa % 2 == 0 for e, kappa in LG.prongs_list()):
            return admcycles.admcycles.prodtautclass(stgraph, terms=[])

        # Check whether the prodtautclass will be zero
        # If any level has extra freedom of scaling, the class will be zero on
        # moduli space of curves.
        if any(self.level(l).zeroStratumClass()
               for l in range(self.codim + 1)):
            if rearrange_markings:
                stgraph = self._G.relabel(self._G.standard_markings(), tidyup=False).LG.stgraph
            return admcycles.admcycles.prodtautclass(stgraph, terms=[])

        # To handle the case that the stratum is connected
        if len(LG.genera) == 1:
            adm_psis = admcycles.admcycles.decstratum(stgraph,
                                                      psi=self.leg_dict)
            adm_psis_taut = admcycles.admcycles.tautclass([adm_psis])

            # To handle the case that the genus is 0 and there are no
            # simple poles.
            if LG.genera[0] == 0 and self.X.sim_pole_list == []:
                stratum_class = AdditiveGenerator(self.X, ((), 0)).as_taut()
                stratum_class = stratum_class.to_prodtautclass().pushforward()
                protaut = [adm_psis_taut * stratum_class]
                result = admcycles.admcycles.prodtautclass(stgraph,
                                                           protaut=protaut)
                return result

            # To handle the case that there are only two poles which are a pair
            # of simple poles.
            if (self.X.res_cond == [] or
               len(self.X.eff_pair_res) < len(self.X.pair_res_cond)):
                stratum_class = self.X.smooth_con_stratum_prodtautclass()
                stratum_class = stratum_class.pushforward()
                protaut = [stratum_class * adm_psis_taut]
                result = admcycles.admcycles.prodtautclass(stgraph,
                                                           protaut=protaut)
                return result

            # To resolve residue condition of a pair of simple poles if the
            # residue condition is extra.
            elif (self.X.eff_pair_res != [] and
                  len(self.X.eff_pair_res) == len(self.X.pair_res_cond)):
                Q = self.X.remove_pair_res()
                total = Q.to_prodtautclass_spin().pushforward()
                protaut = [adm_psis_taut * total]
                result = admcycles.admcycles.prodtautclass(stgraph,
                                                           protaut=protaut)
                return result

        # Now it processes to break down the level graph into level
        # stratum and resolve the residue conditions (non simple poles).
        # Then it computes the result by level clutching.
        # The level graph that can come to this stage will have only odd
        # enhancements or the stratum is not connected, i.e. no vertical edges.
        # In the both cases, the spin class depends on the spin class on
        # each level.

        alpha = []
        vertices = []
        for l in range(self.codim + 1):
            psis = self.psis_on_level(l)
            level = self.level(l)
            level = addspin(level)  # level stratum with spin structure

            if (level.eff_pair_res != [] and
               len(level.eff_pair_res) == len(level.pair_res_cond)):
                # To remove residue condition on a pair of simple poles on
                # a disconnected stratum.
                ptc = level.remove_pair_res(psis).to_prodtautclass_spin()
            else:
                T = level.remove_res_cond(psis)
                ptc = ELGT_addspin(T).to_prodtautclass_spin()
            # print(str(ptc0))
            # print(str(ptc1))
            alpha.append(ptc)
            vertices.append(LG.verticesonlevel(LG.internal_level_number(l)))

        if rearrange_markings:  # if we want the labels of stgraph correspond to the markings
            prodtautst = self._G.relabel(self._G.standard_markings(), tidyup=False).LG.stgraph
            prod = admcycles.admcycles.prodtautclass(prodtautst)

        else:
            prod = admcycles.admcycles.prodtautclass(stgraph)

        # To clutch the level stratum classes together
        for l, ptc in enumerate(alpha):
            prod = prod.factor_pullback(vertices[l], ptc)

        return self.stack_factor * prod


def standardise_sig(sig, res_cond=[]):
    r"""
    To standardise the arrangement of entries in the signature, such that it go from low to high and all the 0s
    are placed at the right end. This is just to reduce the required memory and time to compute the stratum classes.
    It returns

            1. the simplified sig (0s removed);
            2. number of 0s;
            3. the leg_dict of rearrangement;
            4. new res_cond

    INPUT:

        - sig (tuple): a tuple of integers
        - res_cond (list, default=[]): the list of res_cond

    EXAMPLES::

        sage: from admcycles.diffstrata.spinstratum import standardise_sig
        sage: standardise_sig((2,4,-1,0,-1), res_cond=[[(0,2),(0,4)]])
        [(-1, -1, 2, 4), 1, {1: 3, 2: 4, 3: 1, 4: 5, 5: 2}, [((0, 0), (0, 1))]]

    """
    res = deepcopy(res_cond)
    new_rc = []
    l = list(set(sig))
    g = Signature(sig).g
    l.sort()
    t = list(sig)
    freq = {x: t.count(x) for x in l}
    counter = deepcopy(freq)
    standard_sig_without_regular = []

    for x in l:
        if x != 0:
            standard_sig_without_regular += freq[x] * [x]
    m = len(standard_sig_without_regular)
    if g == 0 and m < 3:
        standard_sig_without_regular += (3 - m) * [0]

    leg_dict = {}
    for i in range(0, len(sig)):
        if sig[i] == 0:
            p = sum([f for j, f in freq.items() if j != 0])
            leg_dict[i + 1] = p + freq[0] - counter[0] + 1
            counter[0] -= 1
        if sig[i] != 0:
            p = sum([f for j, f in freq.items() if (j != 0 and j < sig[i])])
            leg_dict[i + 1] = p + freq[sig[i]] - counter[sig[i]] + 1
            counter[sig[i]] -= 1
    output = []
    output.append(tuple(standard_sig_without_regular))
    if 0 in l:
        if g == 0 and m < 3:
            output.append(freq[0] - 3 + m)
        else:
            output.append(freq[0])
    else:
        output.append(0)
    output.append(leg_dict)

    for r in res:
        w = []
        for q in r:
            a = (0, leg_dict[q[1] + 1] - 1)
            w.append(a)
        new_rc.append(tuple(w))
    output.append(new_rc)
    return output


@cached_function
def smooth_stratum_standard(sig, res_cond=()):
    r"""
    This function is to compute the spin stratum class of a standardised
    signature i.e. it is ordered and has no 0s (or all are 0s), also the
    stratum has only one component. The results are cached.

    EXAMPLES::

        sage: from admcycles.diffstrata.spinstratum import smooth_stratum_standard
        sage: smooth_stratum_standard((-1,-1,4))
        [2,
         3,
         2,
         (0, 0, 4, 0, -11/2, -12, 0, -7/2, 0, 1/2, 35/2, 4, 0, -12, 0, 0, 0, -2, 19/2, 0, 7/2, 0, 0, 0, 2, 0, -7/2, 0, 2, -2, 5/2, -13/2, 9/2, 8, 0, 0, 0, 17/2, -11, 2, -2, -9/2, 0, 2)]


    """
    assert (sum([x for x in sig]) == -2 and len(sig) == 3) or all(j == 0 for j in sig) or all(j != 0 for j in sig)

    if res_cond != ():
        rc = []
        for q in res_cond:
            res_single_cond = [t for t in q]
            rc.append(res_single_cond)
        X = SpinStratum([Signature(sig)], rc)
    else:
        X = SpinStratum([Signature(sig)])

    # LG = X.smooth_LG.LG
    # stgraph = LG.stgraph
    g = X._sig_list[0].g
    n = X._sig_list[0].n
    r = 0
    if any(x < 0 for x in sig):
        r = len(res_cond) + 1
    k = g + r - 1
    # The codim we are dealing with for computation of stratum class.

    # length=len(admcycles.admcycles.generating_indices(g,n,k))
    check_basic = is_base_case(sig, g, n, k)

    # To check whether the stratum class is the base cases
    if check_basic != []:
        return check_basic

    # Solve the pullback equations
    ListM = Pullback_Matrices(g, n, r)
    prodtaut_list = bdry_pullback_classify(X)
    tensorTaut_list = []
    for i, u in enumerate(prodtaut_list):
        if u != []:
            t = sum(u)
            v = vector(t.totensorTautbasis(k, vecout=True))
            tensorTaut_list.append(v)
        else:
            length = len(ListM[i][0])
            v = vector(QQ, length)
            tensorTaut_list.append(v)

    sol_inTautbasis = SolveMultLin(ListM, tensorTaut_list)
    return [g, n, k, sol_inTautbasis]

# To give the stratum classes with spin for the easy cases.


def is_base_case(sig, g, n, k):
    if g == 0:
        if all(q % 2 == 0 for q in sig):
            return [g, n, k, vector([QQ(1)])]

        # case of (0,-1,-1)
        if n == 3:
            return [g, n, k, vector([-QQ(1)])]

        # case of (2k,-2k,-1,-1)
        if n == 4:
            result = admcycles.admcycles.psiclass(1, 0, 4).basis_vector()
            return [g, n, k, result]

    if g == 1:
        if all(t == 0 for t in sig):
            return [g, n, k, vector([-QQ(1)])]

    if g == 2 and n == 1:
        return [g, n, k, vector([QQ(1) / 2, -QQ(7) / 2, QQ(1) / 2])]

    return []


def bdry_pullback_classify(stratum):
    X = stratum
    sig = X._sig_list[0].sig
    g = X._sig_list[0].g
    n = X._sig_list[0].n
    # r = 0
    # if any(m < 0 for m in sig):
    #    r = len(X.res_cond) + 1
    # k = g + r - 1
    Agraphs = admcycles.admcycles.list_strata(g, n, 1)
    if g != 0 and n >= 3:
        Agraphs = [G for i, G in enumerate(Agraphs) if i != 0]
    else:
        G = Agraphs[0].copy()
        G.tidy_up()
        Agraphs[0] = G

    bddiv_codim1_LG_prodtautclass_list = [[] for G in Agraphs]
    bic_stgraph = [bic.relabel(bic.standard_markings(), tidyup=False) for bic in X.bics]

    for j, G in enumerate(Agraphs):
        sig0 = [sig[i - 1] for i in G._legs[0] if i < n + 1]
        node_sing = 2 * G._genera[0] - 2 - sum(sig0)

        if node_sing == -1:
            sig0.append(-1)
            sig1 = [sig[i - 1] for i in G._legs[1] if i < n + 1] + [-1]
            new_sig_list = [sig0, sig1]
            splt_st0 = GeneralisedStratum([Signature(tuple(new_sig_list[0]))])
            splt_st1 = GeneralisedStratum([Signature(tuple(new_sig_list[1]))])
            elgt0 = splt_st0.additive_generator(((), 0)).as_taut()
            elgt1 = splt_st1.additive_generator(((), 0)).as_taut()
            st0_spin = ELGT_addspin(elgt0).to_prodtautclass_spin()
            st1_spin = ELGT_addspin(elgt1).to_prodtautclass_spin()
            prodtaut = admcycles.admcycles.prodtautclass(G)

            # Note that for a horizontal one edge graph, the splitting of the
            # two vertices will give rise to extra symplectic basis and thus
            # if sum of spins are odd (even), then the horizontal graph
            # has spin even(odd).
            clutch = prodtaut.factor_pullback([0], st0_spin)
            clutch = clutch.factor_pullback([1], -st1_spin)
            bddiv_codim1_LG_prodtautclass_list[j].append(clutch)

        else:
            if len(G._genera) == 1:
                sig_new = ([sig[i - 1] for i in G._legs[0] if i < n + 1] +
                           [-1, -1])
                Y = SpinStratum([Signature(tuple(sig_new))],
                                res_cond=[[(0, n), (0, n + 1)]])
                elgt = Y.additive_generator(((), 0)).as_taut()
                prodtaut = admcycles.admcycles.prodtautclass(G)

                stratum_class = ELGT_addspin(elgt).to_prodtautclass_spin()
                Intersect = prodtaut.factor_pullback([0], stratum_class)
                # print(str(Intersect))
                bddiv_codim1_LG_prodtautclass_list[j].append(Intersect)
            for num, bic in enumerate(bic_stgraph):
                stgraph = bic.LG.stgraph
                a = G._edges[0][0]
                b = G._edges[0][1]

                Astructure_list = admcycles.admcycles.Astructures(stgraph, G)

                if Astructure_list != []:
                    # print(str(Astructure_list))
                    AG = AG_addspin(X.additive_generator(((num,), 0)))
                    prodtaut_temp = AG.to_prodtautclass_spin()
                    # print(str(prodtaut_temp))
                    ell = bic.ell
                    for w in Astructure_list:
                        G_no_edge = StableGraph(G._genera, G._legs, [])
                        image_half_edges = [(w[1][e[0]], w[1][e[1]])
                                            for e in G._edges]
                        st_edges_new = [e for e in stgraph._edges if
                                        (e not in image_half_edges) and
                                        ((e[1], e[0]) not in image_half_edges)]
                        st_new = StableGraph(stgraph._genera,
                                             stgraph._legs, st_edges_new)
                        prodtaut = admcycles.admcycles.prodtautclass(st_new, prodtaut_temp.terms)
                        # print(str(G_no_edge)+"\n"+str(st_new))
                        E_temp = prodtaut.partial_pushforward(G_no_edge,
                                                              w[0], w[1])
                        E = admcycles.admcycles.prodtautclass(G, E_temp.terms)
                        # print(str(E_temp)+"\n" +str(E))
                        kappa_e = bic.LG.prong((w[1][a], w[1][b]))
                        Intersect = QQ(ell / kappa_e) * E
                        bddiv_codim1_LG_prodtautclass_list[j].append(Intersect)

    return bddiv_codim1_LG_prodtautclass_list


# To generate a list of matrices of the clutching pullback maps.
@cached_function
def Pullback_Matrices(g, n, r):
    k = g + r - 1
    L1 = admcycles.admcycles.tautgens(g, n, k)
    Listgen1 = admcycles.admcycles.generating_indices(g, n, k)

    graph_list = admcycles.admcycles.list_strata(g, n, 1)
    if g != 0 and n >= 3:
        graph_list = [G for i, G in enumerate(graph_list) if i != 0]

    M_list = []
    for G in graph_list:
        M0 = []
        if len(G._genera) == 1:
            G = G.copy()
            G.tidy_up()
        for i in Listgen1:
            T = G.boundary_pullback(L1[i])
            M0 = M0 + [T.totensorTautbasis(k, vecout=True)]
        M = matrix(M0)
        M_list.append(M)
    return M_list


#######################################################################
#######################################################################
# Define the spin version ELGTautClass ################################
#######################################################################


class ELGT_with_spin (ELGTautClass):

    def __init__(self, X, psi_list):
        assert isinstance(X, SpinStratum)
        super().__init__(X, psi_list)
        self.X = X

    def __repr__(self):
        return "ELGT_SpinClass(X=%r,psi_list=%r)"\
            % (self._X, self._psi_list)

    def __str__(self):
        str = "ELGT_SpinClass on %s\n" % self._X
        for coeff, psi in self._psi_list:
            str += "%s * %s + \n" % (coeff, psi)
        return str

    def to_prodtautclass_spin(self):
        G = self.X.smooth_LG
        # print(str(G))
        stgraph = G.LG.stgraph
        total = admcycles.admcycles.prodtautclass(stgraph, terms=[])
        for c, AG in self._psi_list:
            legdictlen = len(AG._G.dmp)
            ptc = AG.to_prodtautclass_spin()
            vertex_map = {}
            for v, _ in enumerate(G.LG.genera):
                mp_on_stratum = G.dmp[G.LG.legs[v][0]]
                l_AG = AG._G.dmp_inv[mp_on_stratum]
                LG = AG._G.LG
                v_AG = LG.vertex(l_AG)
                UG_v = LG.UG_vertex(v_AG)
                T = LG.UG_without_infinity()
                T = T.connected_component_containing_vertex(UG_v)
                for w, g, kind in T:
                    if kind != 'LG':
                        continue
                    vertex_map[w] = v
            leg_map = {i: i for i in range(1, legdictlen + 1)}
            # print(str(stgraph)+"\n"+ str(leg_map)+ "\n"+ str(vertex_map)+ "\n" + str(self)+
            # "\n" + str(ptc))
            pf = ptc.partial_pushforward(stgraph, vertex_map, leg_map)
            # print(str(pf) + "\n" + str(ptc)+ "\n" + str(c))
            total += c * pf
        return total


#############################################################################
# Some useful functions #####################################################
#############################################################################


# turn a ELGTautClass to spin version

def ELGT_addspin(elgt):
    X = SpinStratum(elgt._X._sig_list, elgt._X._res_cond)
    psi_list = [(c, AG_addspin(AG)) for c, AG in elgt._psi_list]
    return ELGT_with_spin(X, psi_list)


# turn a AdditiveGenerator to spin version

def AG_addspin(ag):
    assert isinstance(ag, AdditiveGenerator)
    X = addspin(ag._X)
    return AG_with_spin(X, ag._enh_profile, ag._leg_dict)


# turn a stratum to spin version

def addspin(stratum):
    assert isinstance(stratum, GeneralisedStratum)
    return SpinStratum(stratum._sig_list, stratum._res_cond)


def save_spin_stratum():
    cachedict = dict(smooth_stratum_standard.cache)
    sage.misc.persist.save(cachedict, 'spin_stratum_cache')


def load_spin_stratum():
    cachedict = sage.misc.persist.load('spin_stratum_cache.sobj')
    for a, b in cachedict.items():
        smooth_stratum_standard.set_cache(b, *a[0])



r"""
Generators of tautological ring

Each generator is called a stratum and is encoded in a class class:`Graph`. A
stratum is called *pure* if it has neither kappa nor psi marking (ie a stable
graph with no decoration).
"""
from copy import copy
import itertools

from sage.misc.cachefunc import cached_function
from sage.misc.lazy_attribute import lazy_attribute

from sage.rings.all import PolynomialRing, ZZ
from sage.combinat.all import IntegerVectors, Partitions
from sage.functions.other import floor
from sage.matrix.constructor import matrix

from .utils import aut
from .moduli import MODULI_SM, MODULI_RT, MODULI_CT, MODULI_ST, MODULI_SMALL, dim_form


R = PolynomialRing(ZZ, 1, order='lex', names=('X',))
X = R.gen()


class Graph:
    r"""
    A decorated graph representing a tautological class.

    The graph is encoded as a matrix where

    - each row after the first one corresponds to a vertex
    - each column after the first one corresponds to an edge or leg
    - the first column gives the genera of the vertices
    - the first row gives the markings on the legs
    - the other cells are
      - 0 if the vertex and edge/leg are not incident
      - 1 if the vertex and edge/leg are incident once
      - 2 if the edge is a loop at the vertex
    - entries with polynomials in X describe kappa/psi decorations:
    - in the first column, each X^n term corresponds to a kappa_n at that
      vertex
    - in other locations, each X term corresponds to a psi at that half-edge
    - at loops, 2 + aX + bX^2 corresponds to having psi^a on one side of the
      loop and psi^b on the other

    EXAMPLES::

        sage: from admcycles.DR.graph import Graph, X
        sage: gr = Graph(matrix(2, 3, [-1, 1, 2, 1, X + 1, 1]))
        sage: gr
        [   -1     1     2]
        [    1 X + 1     1]
        sage: gr.num_vertices()
        1
        sage: gr.num_edges()
        2
    """
    def __init__(self, M=None, genus_list=None):
        if M:
            self.M = copy(M)
        elif genus_list:
            self.M = matrix(R, len(genus_list) + 1,
                            1, [-1] + genus_list)
        else:
            self.M = matrix(R, 1, 1, -1)

    def __repr__(self):
        return repr(self.M)

    @lazy_attribute
    def degree_vec(self):
        r"""
        Return the tuple of degrees of vertices.

        EXAMPLES::

            sage: from admcycles.DR.graph import Graph, X
            sage: gr = Graph(matrix(2, 3, [-1, 1, 2, 1, X + 1, 1]))
            sage: gr.degree_vec
            (2,)
        """
        res = [0 for i in range(1, self.M.nrows())]
        for i in range(1, self.M.nrows()):
            for j in range(1, self.M.ncols()):
                res[i - 1] += self.M[i, j][0]
        return tuple(res)

    @lazy_attribute
    def target_parity(self):
        ans = 0
        for i in range(1, self.M.nrows()):
            local_parity = 1 + self.M[i, 0][0]
            for j in range(1, self.M.ncols()):
                local_parity += self.M[i, j][1] + self.M[i, j][2]
            for j in range(1, self.M[i, 0].degree() + 1):
                local_parity += j * self.M[i, 0][j]
            local_parity %= 2
            ans += (local_parity << (i - 1))
        return ans

    def degree(self, i):
        r"""
        Return the degree of vertex ``i``.
        """
        return self.degree_vec[i - 1]

    def num_vertices(self):
        return self.M.nrows() - 1

    def num_edges(self):
        return self.M.ncols() - 1

    def h1(self):
        return self.M.ncols() - self.M.nrows() + 1

    def add_vertex(self, g):
        self.M = self.M.stack(matrix(1, self.M.ncols()))
        self.M[-1, 0] = g

    def add_edge(self, i1, i2, marking=0):
        self.M = self.M.augment(matrix(self.M.nrows(), 1))
        self.M[0, -1] = marking
        if i1 > 0:
            self.M[i1, -1] += 1
        if i2 > 0:
            self.M[i2, -1] += 1

    def del_vertex(self, i):
        self.M = self.M[:i].stack(self.M[(i + 1):])

    def del_edge(self, i):
        if i == self.num_edges():
            self.M = self.M[0:, :i]
        else:
            self.M = self.M[0:, :i].augment(self.M[0:, (i + 1):])

    def split_vertex(self, i, row1, row2):
        self.M = self.M.stack(matrix(2, self.M.ncols(), row1 + row2))
        self.add_edge(self.M.nrows() - 2,
                      self.M.nrows() - 1)
        self.del_vertex(i)

    def replace_vertex_with_graph(self, i, G):
        nv = self.num_vertices()
        ne = self.num_edges()
        # i should have degree d, there should be no classes near i, and G should have genus equal to the genus of i
        hedge_list_ones = []
        hedge_list_others = []
        unsym_cols = [-1]
        mark_nr = -1
        for k in range(1, G.M.ncols()):
            if G.M[0, k] > 0:
                if G.M[0, k] == 1:
                    mark_nr += 1
                unsym_cols.append(G.M[0, k] + mark_nr)
            else:
                unsym_cols.append(0)
        for k in range(1, self.M.ncols()):
            for j in range(self.M[i, k][0]):
                if self.M[0, k] == 1:
                    hedge_list_ones.append(k)
                else:
                    hedge_list_others.append(k)
        hedge_list = hedge_list_ones + hedge_list_others
        self.del_vertex(i)
        for j in range(G.num_edges() - len(hedge_list)):
            self.add_edge(0, 0)
        for j in range(G.num_vertices()):
            self.add_vertex(G.M[j + 1, 0])
        col = ne + 1
        for k in range(1, G.M.ncols()):
            if G.M[0, k] > 0:
                mark = ZZ(unsym_cols[k])
                for j in range(G.num_vertices()):
                    if self.M[nv + j, hedge_list[mark - 1]] == 0:
                        self.M[nv + j, hedge_list[mark - 1]] = G.M[j + 1, k]
                    elif G.M[j + 1, k] != 0:
                        a = self.M[nv + j, hedge_list[mark - 1]][1]
                        b = G.M[j + 1, k][1]
                        self.M[nv + j, hedge_list[mark - 1]] = 2 + \
                            max(a, b) * X + min(a, b) * X**2
            else:
                for j in range(G.num_vertices()):
                    self.M[nv + j, col] = G.M[j + 1, k]
                col += 1

    def compute_invariant(self):
        nr, nc = self.M.nrows(), self.M.ncols()
        self.invariant = [[self.M[i, 0], [], [], [
            [] for j in range(1, nr)]] for i in range(1, nr)]
        for k in range(1, nc):
            L = [i for i in range(1, nr)
                 if self.M[i, k] != 0]
            if len(L) == 1:
                if self.M[0, k] != 0:
                    self.invariant[L[0] -
                                   1][2].append((self.M[0, k], self.M[L[0], k]))
                else:
                    self.invariant[L[0] - 1][1].append(self.M[L[0], k])
            else:
                self.invariant[L[0] - 1][3][L[1] - 1].append((self.M[L[0], k],
                                                              self.M[L[1], k]))
                self.invariant[L[1] - 1][3][L[0] - 1].append((self.M[L[1], k],
                                                              self.M[L[0], k]))
        for i in range(1, nr):
            self.invariant[i - 1][3] = [term for term in self.invariant[i - 1][3]
                                        if len(term)]
            for k in range(len(self.invariant[i - 1][3])):
                self.invariant[i - 1][3][k].sort()
                self.invariant[i -
                               1][3][k] = tuple(self.invariant[i - 1][3][k])
            self.invariant[i - 1][3].sort()
            self.invariant[i - 1][3] = tuple(self.invariant[i - 1][3])
            self.invariant[i - 1][2].sort()
            self.invariant[i - 1][2] = tuple(self.invariant[i - 1][2])
            self.invariant[i - 1][1].sort()
            self.invariant[i - 1][1] = tuple(self.invariant[i - 1][1])
            self.invariant[i - 1] = tuple(self.invariant[i - 1])
        vertex_invariants = [[i, self.invariant[i - 1]]
                             for i in range(1, nr)]
        self.invariant.sort()
        self.invariant = tuple(self.invariant)
        vertex_invariants.sort(key=lambda x: x[1])
        self.vertex_groupings = []
        for i in range(nr - 1):
            if i == 0 or vertex_invariants[i][1] != vertex_invariants[i - 1][1]:
                self.vertex_groupings.append([])
            self.vertex_groupings[-1].append(
                vertex_invariants[i][0])

    def purify(self):
        for i in range(self.M.nrows()):
            for j in range(self.M.ncols()):
                self.M[i, j] = R(self.M[i, j][0])

    def contract(self, i, vlist, elist):
        # assumes graph is undecorated
        if self.M[0, i] != 0:
            print("ERROR: cannot contract a marking")
            return
        S = [row for row in range(
            1, self.M.nrows()) if self.M[row, i] != 0]
        if len(S) == 1:
            self.M[S[0], 0] += 1
            self.del_edge(i)
            elist = elist[:(i - 1)] + elist[i:]
        else:
            self.del_edge(i)
            elist = elist[:(i - 1)] + elist[i:]
            self.add_vertex(0)
            self.M[-1] += self.M[S[0]]
            self.M[-1] += self.M[S[1]]
            self.del_vertex(S[1])
            self.del_vertex(S[0])
            vlist = vlist[:(S[0] - 1)] + vlist[S[0]:(S[1] - 1)] + \
                vlist[S[1]:] + [vlist[S[0] - 1] + vlist[S[1] - 1]]
        return vlist, elist


def graph_isomorphic(G1, G2):
    r"""
    Return whether ``G1`` and ``G2`` are isomorphic.
    """
    if G1.invariant != G2.invariant:
        return False

    M1 = G1.M
    M2 = G2.M
    group1 = G1.vertex_groupings
    group2 = G2.vertex_groupings

    nr = M1.nrows()
    nc = M1.ncols()
    for sigma_data in itertools.product(*(itertools.permutations(range(len(group))) for group in group1)):
        sigma = [0] * (nr - 1)
        for i in range(len(group1)):
            for j in range(len(group1[i])):
                sigma[group1[i][j] - 1] = group2[i][sigma_data[i][j]]
        good = True
        for i in range(1, nr):
            ii = sigma[i - 1]
            for j in range(1, i):
                jj = sigma[j - 1]
                L1 = []
                for k in range(1, nc):
                    if M1[i, k] != 0 and M1[j, k] != 0:
                        L1.append([M1[i, k], M1[j, k]])
                L1.sort()
                L2 = []
                for k in range(1, nc):
                    if M2[ii, k] != 0 and M2[jj, k] != 0:
                        L2.append([M2[ii, k], M2[jj, k]])
                L2.sort()
                if L1 != L2:
                    good = False
                    break
            if good is False:
                break
        if good:
            return True
    return False


def graph_count_automorphisms(G, vertex_orbits=False):
    return count_automorphisms(G.M, G.vertex_groupings, vertex_orbits)


def count_automorphisms(M, grouping, vertex_orbits=False):
    """
    INPUT:

    M: adjacency matrix
    grouping: list of lists
    vertex_orbits: boolean
    """
    nr, nc = M.nrows(), M.ncols()
    count = ZZ.zero()
    if vertex_orbits:
        isom_list = []
    for sigma_data in itertools.product(*(itertools.permutations(range(len(group))) for group in grouping)):
        sigma = [0 for i in range(nr - 1)]
        for i in range(len(grouping)):
            for j in range(len(grouping[i])):
                sigma[grouping[i][j] - 1] = grouping[i][sigma_data[i][j]]
        good = True
        for i in range(1, nr):
            ii = sigma[i - 1]
            for j in range(1, i):
                jj = sigma[j - 1]
                L1 = []
                for k in range(1, nc):
                    if M[i, k] != 0 and M[j, k] != 0:
                        L1.append([M[i, k], M[j, k]])
                L1.sort()
                L2 = []
                for k in range(1, nc):
                    if M[ii, k] != 0 and M[jj, k] != 0:
                        L2.append([M[ii, k], M[jj, k]])
                L2.sort()
                if L1 != L2:
                    good = False
                    break
            if not good:
                break
        if good:
            count += 1
            if vertex_orbits:
                isom_list.append(sigma)

    if vertex_orbits:
        orbit_list = []
        vertices_used = []
        while len(vertices_used) < nr - 1:
            i = next(ii for ii in range(1, nr) if ii not in vertices_used)
            orbit = []
            for sigma in isom_list:
                if sigma[i - 1] not in orbit:
                    orbit.append(sigma[i - 1])
                    vertices_used.append(sigma[i - 1])
            orbit.sort()
            orbit_list.append(orbit)
        return orbit_list

    for i in range(1, nr):
        for k in range(1, nc):
            if M[i, k][0] == 2 and M[i, k][1] == M[i, k][2]:
                count *= 2
        L = []
        for k in range(1, nc):
            if M[i, k] != 0:
                if sum(1 for j in range(1, nr) if M[j, k] != 0) == 1:
                    L.append([M[0, k], M[i, k]])
        count *= aut(L)

        for j in range(1, i):
            L = []
            for k in range(1, nc):
                if M[i, k] != 0 and M[j, k] != 0:
                    L.append([M[i, k], M[j, k]])
            count *= aut(L)
    return count


def graph_list_isomorphisms(G1, G2, only_one=False):
    r"""
    Return the list of isomorphisms from ``G1`` to ``G2``.
    """
    # Warning: does not count loops!
    # If this is too slow, we can probably improve by caching a list of automorphisms and applying those to the first isom found.
    if G1.invariant != G2.invariant:
        return []

    M1 = G1.M
    M2 = G2.M
    group1 = G1.vertex_groupings
    group2 = G2.vertex_groupings

    nr = M1.nrows()
    nc = M2.ncols()
    isom_list = []
    for sigma_data in itertools.product(*(itertools.permutations(range(len(group))) for group in group1)):
        sigma = [0] * (nr - 1)
        for i in range(len(group1)):
            for j in range(len(group1[i])):
                sigma[group1[i][j] - 1] = group2[i][sigma_data[i][j]]
        good = True
        for i in range(1, nr):
            ii = sigma[i - 1]
            for j in range(1, i):
                jj = sigma[j - 1]
                L1 = []
                for k in range(1, nc):
                    if M1[i, k] != 0 and M1[j, k] != 0:
                        L1.append([M1[i, k], M1[j, k]])
                L1.sort()
                L2 = []
                for k in range(1, nc):
                    if M2[ii, k] != 0 and M2[jj, k] != 0:
                        L2.append([M2[ii, k], M2[jj, k]])
                L2.sort()
                if L1 != L2:
                    good = False
                    break
            if not good:
                break
        if good:
            cols1 = [[M1[i, j] for i in range(nr)] for j in range(1, nc)]
            cols2 = [[M2[0, j]] + [M2[sigma[i - 1], j] for i in range(1, nr)]
                     for j in range(1, nc)]
            edge_group1 = []
            edge_group2 = []
            used1 = []
            for j in range(1, nc):
                if j not in used1:
                    edge_group1.append([])
                    edge_group2.append([])
                    for k in range(1, nc):
                        if cols1[k - 1] == cols1[j - 1]:
                            edge_group1[-1].append(k)
                            used1.append(k)
                        if cols2[k - 1] == cols1[j - 1]:
                            edge_group2[-1].append(k)
            for edge_sigma_data in itertools.product(*(itertools.permutations(range(len(edge_group))) for edge_group in edge_group1)):
                edge_sigma = [0 for i in range(nc - 1)]
                for i in range(len(edge_group1)):
                    for j in range(len(edge_group1[i])):
                        edge_sigma[edge_group1[i][j] -
                                   1] = edge_group2[i][edge_sigma_data[i][j]]
                isom_list.append([sigma, edge_sigma])
                if only_one:
                    return isom_list
    return isom_list


def degenerate(G_list, moduli_type=MODULI_ST):
    r"""
    Return the list of degenerations of the graphs in ``G_list`` up to
    isomorphisms.
    """
    mod_size = moduli_type + 1
    if moduli_type == MODULI_SMALL:
        mod_size = MODULI_SM + 1
    G_list_new = [[] for i in range(mod_size)]
    for which_type in range(mod_size):
        for G in G_list[which_type]:
            for i in range(1, G.num_vertices() + 1):
                row = list(G.M[i])
                m = row[0] + sum(row)
                if m < 4:
                    continue
                row1 = [0 for j in range(len(row))]
                while [2 * x for x in row1] <= row:
                    if row1[0] == 1 and moduli_type <= MODULI_RT:
                        break
                    if row1[0] + sum(row1) >= 2 and row1[0] + sum(row1) <= m - 2:
                        row2 = [row[j] - row1[j] for j in range(len(row))]
                        G_copy = Graph(G.M)
                        G_copy.split_vertex(i, row1, row2)
                        new_type = which_type
                        if new_type == MODULI_SM:
                            new_type = MODULI_RT
                        if new_type == MODULI_RT and row1[0] > 0:
                            new_type = MODULI_CT
                        G_list_new[new_type].append(G_copy)
                    row1[-1] += 1
                    for j in range(1, len(row)):
                        if row1[-j] <= row[-j]:
                            break
                        row1[-j] = 0
                        row1[-j - 1] += 1
    for i in range(mod_size):
        G_list_new[i] = remove_isomorphic(G_list_new[i])
    return G_list_new


def decorate(G_list, r, moduli_type=MODULI_ST):
    r"""
    Add decorations to the graphs in ``G_list``.
    """
    mod_size = moduli_type + 1
    if moduli_type == MODULI_SMALL:
        mod_size = MODULI_SM + 1
    G_list_new = [[] for i in range(mod_size)]
    for which_type in range(mod_size):
        for G in G_list[which_type]:
            G_deco = [[] for i in range(mod_size)]
            nr, nc = G.M.nrows(), G.M.ncols()
            two_list = []
            one_list = []
            for i in range(1, nr):
                for j in range(1, nc):
                    if G.M[i, j] == 2:
                        two_list.append([i, j])
                    elif G.M[i, j] == 1:
                        one_list.append([i, j])
            a = nr - 1
            b = len(two_list)
            c = len(one_list)
            dims = [[dim_form(G.M[i + 1, 0][0], G.degree(i + 1), mod_type)
                     for i in range(a)] for mod_type in range(mod_size)]
            for vec in IntegerVectors(r, a + b + c):
                new_type = which_type
                if moduli_type > MODULI_SMALL:
                    test_dims = vec[:a]
                    for i in range(b):
                        test_dims[two_list[i][0] - 1] += vec[a + i]
                    for i in range(c):
                        test_dims[one_list[i][0] - 1] += vec[a + b + i]
                    for mod_type in range(which_type, mod_size):
                        for i in range(a):
                            if test_dims[i] > dims[mod_type][i]:
                                new_type = mod_type + 1
                                break
                    if new_type > moduli_type:
                        continue
                S_list = []
                for i in range(a):
                    S_list.append(Partitions(vec[i]))
                for i in range(a, a + b):
                    S_list.append(
                        [[vec[i] - j, j] for j in range(floor(vec[i] / ZZ(2) + 1))])
                S = itertools.product(*S_list)
                for vec2 in S:
                    G_copy = Graph(G.M)
                    for i in range(a):
                        for j in vec2[i]:
                            G_copy.M[i + 1, 0] += X**j
                    for i in range(a, a + b):
                        G_copy.M[two_list[i - a][0], two_list[i - a]
                                 [1]] += vec2[i][0] * X + vec2[i][1] * X**2
                    for i in range(c):
                        G_copy.M[one_list[i][0],
                                 one_list[i][1]] += vec[i + a + b] * X
                    G_deco[new_type].append(G_copy)
            for mod_type in range(mod_size):
                G_list_new[mod_type] += remove_isomorphic(G_deco[mod_type])
    return G_list_new


def remove_isomorphic(G_list):
    r"""
    Return a list of isomorphism class representatives of the graphs in ``G_list``.
    """
    G_list_new = []
    inv_dict = {}
    count = 0
    for G1 in G_list:
        G1.compute_invariant()
        if G1.invariant not in inv_dict:
            inv_dict[G1.invariant] = []
        good = True
        for i in inv_dict[G1.invariant]:
            if graph_isomorphic(G1, G_list_new[i]):
                good = False
                break
        if good:
            G_list_new.append(G1)
            inv_dict[G1.invariant].append(count)
            count += 1
    return G_list_new


def num_strata(g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the number of strata in given genus, rank, markings and moduli type.

    EXAMPLES::

        sage: from admcycles.DR.moduli import MODULI_SM, MODULI_CT, MODULI_RT, MODULI_ST
        sage: from admcycles.DR.graph import num_strata
        sage: for r in range(4):
        ....:     print('r={}: {} {} {} {}'.format(r,
        ....:           num_strata(2, r, (1, 1), MODULI_SM),
        ....:           num_strata(2, r, (1, 1), MODULI_RT),
        ....:           num_strata(2, r, (1, 1), MODULI_CT),
        ....:           num_strata(2, r, (1, 1), MODULI_ST)))
        r=0: 1 1 1 1
        r=1: 2 3 5 6
        r=2: 0 7 16 28
        r=3: 0 0 38 113
    """
    return len(all_strata(g, r, markings, moduli_type))


def num_pure_strata(g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the number of pure strata in given genus, rank, markings and moduli type.

    EXAMPLES::

        sage: from admcycles.DR.moduli import MODULI_SM, MODULI_CT, MODULI_RT, MODULI_ST
        sage: from admcycles.DR.graph import num_pure_strata
        sage: for r in range(4):
        ....:     print('r={}: {} {} {} {}'.format(r,
        ....:           num_pure_strata(2, r, (1, 1), MODULI_SM),
        ....:           num_pure_strata(2, r, (1, 1), MODULI_RT),
        ....:           num_pure_strata(2, r, (1, 1), MODULI_CT),
        ....:           num_pure_strata(2, r, (1, 1), MODULI_ST)))
        r=0: 1 1 1 1
        r=1: 0 1 3 4
        r=2: 0 0 3 10
        r=3: 0 0 2 19
    """
    return len(all_pure_strata(g, r, markings, moduli_type))


def single_stratum(num, g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the ``num``-th stratum.
    """
    return all_strata(g, r, markings, moduli_type)[num]


def single_pure_stratum(num, g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the ``num``-th pure stratum.
    """
    return all_pure_strata(g, r, markings, moduli_type)[num]


@cached_function
def autom_count(num, g, r, markings=(), moduli_type=MODULI_ST):
    return graph_count_automorphisms(single_stratum(num, g, r, markings, moduli_type))


@cached_function
def pure_strata_autom_count(num, g, r, markings=(), moduli_type=MODULI_ST):
    return graph_count_automorphisms(single_pure_stratum(num, g, r, markings, moduli_type))


@cached_function
def automorphism_cosets(num, g, r, markings=(), moduli_type=MODULI_ST):
    G = single_stratum(num, g, r, markings, moduli_type)
    pureG = Graph(G.M)
    pureG.purify()
    pureG.compute_invariant()
    pure_auts = graph_list_isomorphisms(pureG, pureG)
    num_pure = len(pure_auts)
    impure_auts = graph_list_isomorphisms(G, G)
    num_impure = len(impure_auts)
    chosen_auts = []
    used_auts = []
    v = G.num_vertices()
    e = G.num_edges()
    for i in range(num_pure):
        if i not in used_auts:
            chosen_auts.append(pure_auts[i])
            for g in impure_auts:
                sigma = [[pure_auts[i][0][g[0][k] - 1]
                          for k in range(v)], [pure_auts[i][1][g[1][k] - 1] for k in range(e)]]
                for ii in range(num_pure):
                    if pure_auts[ii] == sigma:
                        used_auts.append(ii)
                        break
    return [num_impure, chosen_auts]


@cached_function
def unpurify_map(g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return a dictionary which maps pure strata to list of strata.

    EXAMPLES::

        sage: from admcycles.DR.graph import unpurify_map
        sage: unpurify_map(2, 2)
        {(0, 0): [0, 1], (1, 0): [2, 3], (1, 1): [4, 5], (2, 0): [6], (2, 1): [7]}
    """
    unpurify = {}
    pure_strata = [all_pure_strata(g, r0, markings, moduli_type)
                   for r0 in range(r + 1)]
    impure_strata = all_strata(g, r, markings, moduli_type)
    for i, strati in enumerate(impure_strata):
        G = Graph(strati.M)
        G.purify()
        r0 = G.num_edges() - len(markings)
        found = False
        for j in range(len(pure_strata[r0])):
            if G.M == pure_strata[r0][j].M:
                G_key = (r0, j)
                found = True
                break
        assert found, "failed purification"
        if G_key not in unpurify:
            unpurify[G_key] = []
        unpurify[G_key].append(i)
    return unpurify


@cached_function
def all_strata(g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the lists of strata with given genus, degree, markings and moduli.

    EXAMPLES::

        sage: from admcycles.DR.graph import all_strata
        sage: all_strata(2, 1, (1, 1))
        [[   -1     1     1]
         [X + 2     1     1],
         [   -1     1     1]
         [    2 X + 1     1],
         [-1  1  1  0]
         [ 0  1  1  1]
         [ 2  0  0  1],
         [-1  1  1  0]
         [ 1  0  0  1]
         [ 1  1  1  1],
         [-1  1  1  0]
         [ 1  0  1  1]
         [ 1  1  0  1],
         [-1  0  1  1]
         [ 1  2  1  1]]
    """
    mod_size = moduli_type + 1
    if moduli_type == MODULI_SMALL:
        mod_size = MODULI_SM + 1
    big_list = [[] for i in range(mod_size)]
    for loops in range(g + 1):
        if loops == 1 and moduli_type <= MODULI_CT:
            break
        if loops > r:
            break
        for edges in range(r - loops + 1):
            if edges == 1 and moduli_type <= MODULI_SM:
                break
            G = Graph()
            G.add_vertex(g - loops)
            for k in range(loops):
                G.add_edge(1, 1)
            for k in markings:
                G.add_edge(1, 0, k)
            GGG = [[] for i in range(mod_size)]
            if loops == 0:
                if edges == 0:
                    GGG[MODULI_SM] = [G]
                else:
                    GGG[MODULI_RT] = [G]
            else:
                GGG[MODULI_ST] = [G]
            for k in range(edges):
                GGG = degenerate(GGG, moduli_type)
            GGG = decorate(GGG, r - loops - edges, moduli_type)
            for i in range(mod_size):
                big_list[i] += GGG[i]
    combined_list = []
    for i in range(mod_size):
        combined_list += big_list[i]
    return combined_list


@cached_function
def all_pure_strata(g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the lists of pure strata with given genus, degree, markings and moduli.

    EXAMPLES::

        sage: from admcycles.DR.graph import all_pure_strata
        sage: from admcycles.DR.moduli import MODULI_CT
        sage: all_pure_strata(2, 0, (1, 1))
        [[-1  1  1]
         [ 2  1  1]]
        sage: all_pure_strata(2, 1, (1, 1), MODULI_CT)
        [[-1  1  1  0]
         [ 0  1  1  1]
         [ 2  0  0  1],
         [-1  1  1  0]
         [ 1  0  0  1]
         [ 1  1  1  1],
         [-1  1  1  0]
         [ 1  0  1  1]
         [ 1  1  0  1]]
    """
    big_list = [[] for i in range(moduli_type + 1)]
    for loops in range(g + 1):
        if loops == 1 and moduli_type <= MODULI_CT:
            break
        if loops > r:
            break
        for edges in range(r - loops, r - loops + 1):
            if edges >= 1 and moduli_type <= MODULI_SM:
                break
            G = Graph()
            G.add_vertex(g - loops)
            for k in range(loops):
                G.add_edge(1, 1)
            for k in markings:
                G.add_edge(1, 0, k)
            G.compute_invariant()
            GGG = [[] for i in range(moduli_type + 1)]
            if loops == 0:
                if edges == 0:
                    GGG[MODULI_SM] = [G]
                else:
                    GGG[MODULI_RT] = [G]
            else:
                GGG[MODULI_ST] = [G]
            for k in range(edges):
                GGG = degenerate(GGG, moduli_type)
            for i in range(moduli_type + 1):
                big_list[i] += GGG[i]
    combined_list = []
    for i in range(moduli_type + 1):
        combined_list += big_list[i]
    return combined_list


@cached_function
def strata_invariant_lookup(g, r, markings=(), moduli_type=MODULI_ST):
    inv_dict = {}
    L = all_strata(g, r, markings, moduli_type)
    for i in range(len(L)):
        if L[i].invariant not in inv_dict:
            inv_dict[L[i].invariant] = []
        inv_dict[L[i].invariant].append(i)
    return inv_dict


@cached_function
def num_of_stratum(G, g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the index of the graph ``G`` in the list of all strata with given
    genus, rank, markings and moduli type.

    This is the inverse of :func:`single_stratum`.

    EXAMPLES::

        sage: from admcycles.DR.graph import Graph, X, R, num_of_stratum
        sage: G = Graph(matrix(R, 2, 3, [-1, 1, 2, X+1, 1, 1]))
        sage: num_of_stratum(G, 1, 1, (1, 2))
        0
    """
    G.compute_invariant()
    L = all_strata(g, r, markings, moduli_type)
    LL = strata_invariant_lookup(g, r, markings, moduli_type)
    x = LL[G.invariant]

    if len(x) == 1:
        return x[0]
    for i in x:
        if graph_isomorphic(G, L[i]):
            return i

    raise ValueError("wrong Graph with g={}, r={}, markings={}, moduli_type={}\n{}".format(
        g, r, markings, moduli_type, G))


def list_strata(g, r, n=0, moduli_type=MODULI_ST):
    """
    Displays the list of all strata.

    EXAMPLES::

        sage: from admcycles.DR.graph import list_strata
        sage: list_strata(1,1,2)
        generator 0
        [   -1     1     2]
        [X + 1     1     1]
        ------------------------------
        generator 1
        [   -1     1     2]
        [    1 X + 1     1]
        ------------------------------
        generator 2
        [   -1     1     2]
        [    1     1 X + 1]
        ------------------------------
        generator 3
        [-1  1  2  0]
        [ 0  1  1  1]
        [ 1  0  0  1]
        ------------------------------
        generator 4
        [-1  0  1  2]
        [ 0  2  1  1]
        ------------------------------
    """
    L = all_strata(g, r, tuple(range(1, n + 1)), moduli_type)
    for i, Li in enumerate(L):
        print("generator %s" % i)
        print(Li.M)
        print("-" * 30)


@cached_function
def contraction_table(g, r, markings=(), moduli_type=MODULI_ST):
    contraction_dict = {}
    pure_strata = [all_pure_strata(g, r0, markings, moduli_type)
                   for r0 in range(r + 1)]
    for r0 in range(r + 1):
        for ii in range(len(pure_strata[r0])):
            G = pure_strata[r0][ii]
            S = [j for j in range(1, G.M.ncols()) if G.M[0, j] == 0]
            contractions = {}
            for edge_subset in itertools.product(*[[0, 1] for i in range(r0)]):
                key = tuple(i for i in range(
                    r0) if edge_subset[i] == 0)
                A = [S[i] for i in key]
                A.reverse()
                vlist = [[i] for i in range(1, G.M.nrows())]
                elist = list(range(1, G.M.ncols()))
                Gcopy = Graph(G.M)
                for i in A:
                    vlist, elist = Gcopy.contract(i, vlist, elist)
                Gcopy.compute_invariant()
                rnew = r0 - len(A)
                contraction_result = []
                for i in range(len(pure_strata[rnew])):
                    L = graph_list_isomorphisms(
                        pure_strata[rnew][i], Gcopy, True)
                    if L:
                        contraction_result.append((rnew, i))
                        contraction_result.append(L[0])
                        break
                contraction_result.append((vlist, elist))
                contractions[key] = contraction_result

            for edge_assignment in itertools.product(*[[0, 1, 2] for i in range(r0)]):
                if sum(1 for i in edge_assignment if i == 1) > r - r0:
                    continue
                key1 = tuple(i for i in range(
                    r0) if edge_assignment[i] == 0)
                B = [S[i]
                     for i in range(r0) if edge_assignment[i] == 1]
                key2 = tuple(i for i in range(
                    r0) if edge_assignment[i] == 2)
                if key1 > key2:
                    continue
                contract1 = contractions[key1]
                contract2 = contractions[key2]
                dict_key = [contract1[0], contract2[0]]
                dict_entry = [contract1[1], contract2[1]]
                if dict_key[0] > dict_key[1]:
                    dict_key.reverse()
                    dict_entry.reverse()
                    dict_entry = [(r0, ii), B, contract2[2],
                                  contract1[2]] + dict_entry
                else:
                    dict_entry = [(r0, ii), B, contract1[2],
                                  contract2[2]] + dict_entry
                dict_key = tuple(dict_key)
                if dict_key not in contraction_dict:
                    contraction_dict[dict_key] = []
                contraction_dict[dict_key].append(dict_entry)
                if dict_key[0] == dict_key[1]:
                    contraction_dict[dict_key].append(
                        dict_entry[:2] + [dict_entry[3], dict_entry[2], dict_entry[5], dict_entry[4]])
    return contraction_dict



# -*- coding: utf-8 -*-
r"""
Double ramification cycle
"""

import itertools
from copy import copy

from sage.combinat.integer_vector import IntegerVectors
from sage.combinat.combinat import bernoulli_polynomial
from sage.arith.all import factorial
from sage.functions.log import exp
from sage.rings.all import QQ
from sage.rings.power_series_ring import PowerSeriesRing

from admcycles.admcycles import list_strata
from admcycles.stable_graph import StableGraph
from admcycles.double_ramification_cycle import DR_cycle
from .tautological_ring import TautologicalRing

# S.<x0,x1>=PowerSeriesRing(QQ,'x0,x1',default_prec=14)


def graph_sum(g, n, decgraphs=None, globalfact=None, vertterm=None, legterm=None, edgeterm=None, maxdeg=None, deg=None, termsout=False):
    r"""Returns the (possibly mixed-degree) tautological class obtained by summing over graphs gamma,
    inserting vertex-, leg- and edgeterms.

    INPUT:

    - ``decgraphs`` -- list or generator; entries of decgraphs are pairs (gamma,dec) of a StableGraph
      gamma and some additional combinatorial structure dec associated to gamma

    - ``globalfact`` -- function; globalfact(gamma,dec) gets handed the parameters gamma,dec as arguments and gives out a number that is multiplied with the corresponding term in the graph sum; default is 1

    - ``vertterm`` -- function; ``vertterm(gv,nv,maxdeg, **kwargs)`` takes arguments local genus gv and number of legs nv
      and maxdeg gets handed the parameters gamma,dec,v as optional keyworded arguments and gives out a
      tautological class on Mbar_{gv,nv}; the class is assumed to be of degree at most maxdeg,
      if deg is given, the class is exactly of degree deg

    - ``legterm`` -- function; ``legterm(gv,nv,i,maxdeg, **kwargs)`` similar to vertterm, except input is
      gv,nv,i,maxdeg where i is number of marking on Mbar_{gv,nv} associated to leg
      gamma, dec and origleg (number of leg in outer graph) given as keyworded arguments

    - ``edgeterm`` -- function; edgeterm(maxdeg,**kwargs) takes keyworded arguments gamma,dec,e,maxdeg
      it gives a generating series s in x0,x1 such that the insertion at edge
      e=(h0,h1) is given by s(psi_h0, psi_h1)

    - ``termsout`` -- parameter; if termsout=False, return sum of all terms
      if termsout = 'coarse', return tuple of terms, one for each pair (gamma,dec)
      if termsout = 'fine', return tuple of terms, one for each pair (gamma,dec) and each distribution
      of cohomological degrees to vertices and half-edges
    """
    if maxdeg is None:
        maxdeg = 3 * g - 3 + n if deg is None else deg
    if decgraphs is None:
        decgraphs = [(gr, None) for ednum in range(3 * g - 3 + n + 1)
                     for gr in list_strata(g, n, ednum)]
    if globalfact is None:
        def globalfact(a, b):
            return 1
    if vertterm is None:
        def vertterm(gv, nv, maxdeg, **kwargs):
            return TautologicalRing(gv, nv).fundamental_class()
    if legterm is None:
        def legterm(gv, nv, i, maxdeg, **kwargs):
            return TautologicalRing(gv, nv).fundamental_class()
    if edgeterm is None:
        def edgeterm(maxdeg, **kwargs):
            S = PowerSeriesRing(QQ, 'x0,x1', default_prec=maxdeg + 1)
            return S.one()

    termlist = []

    for (gamma, dec) in decgraphs:
        restdeg = maxdeg - len(gamma.edges())
        if restdeg < 0:
            continue

        gammadectermlist = []

        numvert = gamma.numvert()
        gnvect = [(gamma.genera(i), len(gamma.legs(i))) for i in range(numvert)]
        dimvect = [3 * g - 3 + n for (g, n) in gnvect]
        markings = gamma.list_markings()
        vertdic = {l: v for v in range(numvert) for l in gamma.legs(v)}
        indexdic = {l: j + 1 for v in range(numvert) for j, l in enumerate(gamma.legs(v))}
        # ex_dimvect=[dimvect[vertdic[l]] for l in halfedges]  # list of dimensions of spaces adjacent to half-edges

        # Pre-compute all vertex-, leg- and edgeterms
        vterms = {v: vertterm(gnvect[v][0], gnvect[v][1], restdeg, gamma=gamma, dec=dec, v=v) for v in range(numvert)}
        lterms = {i: legterm(gnvect[vertdic[i]][0], gnvect[vertdic[i]][1], indexdic[i],
                             restdeg, gamma=gamma, dec=dec, origleg=i) for i in markings}
        eterms = {e: edgeterm(restdeg, gamma=gamma, dec=dec, e=e) for e in gamma.edges()}
        varlist = {(h0, h1): eterms[h0, h1].parent().gens() for h0, h1 in gamma.edges()}
        eterms = {e: eterms[e].coefficients() for e in eterms}
        varx = {h0: varlist[h0, h1][0] for h0, h1 in gamma.edges()}
        varx.update({h1: varlist[h0, h1][1] for h0, h1 in gamma.edges()})

        if deg is None:
            rdlis = range(restdeg + 1)  # terms of all degrees up to restdeg must be computed
        else:
            rdlis = [deg - len(gamma.edges())]
        for rdeg in rdlis:
            # distribute the remaining degree rdeg to vertices
            for degdist in IntegerVectors(rdeg, numvert, outer=dimvect):
                # now for each vertex, split degree to vertex- and leg/half-edge terms
                vertchoices = [IntegerVectors(degdist[v], len(gamma.legs(v)) + 1) for v in range(numvert)]
                for choice in itertools.product(*vertchoices):
                    vdims = []
                    ldims = {}
                    for v in range(numvert):
                        vdims.append(choice[v][0])
                        ldims.update({l: choice[v][i + 1] for i, l in enumerate(gamma.legs(v))})

                    effvterms = [vterms[v].degree_part(vdims[v]) for v in vterms]
                    efflterms = {i: lterms[i].degree_part(ldims[i]) for i in lterms}
                    for i in efflterms:
                        effvterms[vertdic[i]] *= efflterms[i]  # multiply contributions from legs to vertexterms
                    for h0, h1 in gamma.edges():
                        # TODO: optimization potential here by multiplying kppolys directly
                        gv0, nv0 = gnvect[vertdic[h0]]
                        R0 = TautologicalRing(gv0, nv0)
                        effvterms[vertdic[h0]] *= eterms[(h0, h1)].get(varx[h0]**ldims[h0]
                                                                       * varx[h1]**ldims[h1], 0) * R0.psi(indexdic[h0])**ldims[h0]
                        gv1, nv1 = gnvect[vertdic[h1]]
                        R1 = TautologicalRing(gv1, nv1)
                        effvterms[vertdic[h1]] *= R1.psi(indexdic[h1])**ldims[h1]
                    for t in effvterms:
                        t.simplify()
                    # print(gamma)
                    # print(rdeg)
                    # print(degdist)
                    # print(choice)
                    # print(effvterms)
                    # print(eterms)
                    # print(indexdic)
                    # print('\n')
                    tempres = gamma.boundary_pushforward(effvterms)
                    tempres.simplify()
                    if not tempres.is_empty():
                        tempres *= globalfact(gamma, dec)
                        gammadectermlist.append(tempres)
                    # print(termlist)
        if termsout == 'coarse':
            termlist.append(sum(gammadectermlist))
        else:
            termlist += gammadectermlist
    if termsout:
        return termlist
    else:
        return sum(termlist)

############
#
# Useful functions and examples
#
############

###
# Example 1 : Conjectural graph sum for DR_g(1,-1)
###

# Generating functions for graphs


def DR11_tree_test(gr):
    return gr.vertex(1) == gr.vertex(2)


def DR11_trees(g, maxdeg):
    return [gr for n in range(1, g + 1) for e in range(min(n, maxdeg - n + 1))
            for gr in list_strata(0, 2 + n, e) if DR11_tree_test(gr)]


def DR11_graphs(g, maxdeg=None):
    if maxdeg is None:
        maxdeg = 3 * g - 3 + 2
    result = []
    for gr in DR11_trees(g, maxdeg):
        n = len(gr.list_markings()) - 2
        maxleg = max([max(j + [0]) for j in gr.legs])
        grlist = []
        for gdist in IntegerVectors(g, n, min_part=1):
            genera = copy(gr.genera) + list(gdist)
            legs = copy(gr.legs) + [[j] for j in range(maxleg + 1, maxleg + n + 1)]
            edges = copy(gr.edges) + [(j - maxleg + 2, j) for j in range(maxleg + 1, maxleg + n + 1)]
            grlist.append(StableGraph(genera, legs, edges))
        result += [(gam, None) for gam in grlist]
    removedups(result, lambda a, b: a[0].is_isomorphic(b[0]))
    return result


def removedups(li, comp=None):
    """
    Remove duplicates in a list ``li`` according to a comparison function.

    This works inplace and modifies ``li``.

    EXAMPLES::

        sage: from admcycles.graph_sum import removedups
        sage: L = [4,6,3,2,4,99,1,3,2]
        sage: removedups(L)
        sage: L
        [6, 4, 99, 1, 3, 2]
    """
    if comp is None:
        def comp(a, b):
            return a == b
    n = len(li)
    currn = len(li)
    for i in range(n, -1, -1):
        if any(comp(li[i], li[j]) for j in range(i + 1, currn)):
            li.pop(i)
            currn -= 1

# Global factor = 1/|Aut(Gamma)|


def divbyaut(gamma, dec):
    return QQ(1) / gamma.automorphism_number()

# Vertex- and edgeterms for DR11


def DR11_vterm(gv, nv, maxdeg, **kwargs):
    R = TautologicalRing(gv, nv)
    if gv == 0:
        gamma = kwargs['gamma']
        v = kwargs['v']
        f = R.fundamental_class()
        if 1 not in gamma.legs[v]:
            # we are in genus zero vertex not equal to base
            return -nv * f
        else:
            # we are at the base vertex
            return f
    else:
        return sum([(-1)**j * R.lambdaclass(j) for j in range(maxdeg + 1)])


def DR11_eterm(maxdeg, **kwargs):  # smarter: edterm(maxdeg=None,gamma=None,dec=None,e=None, *args, *kwds)
    S = PowerSeriesRing(QQ, 'x0,x1', default_prec=maxdeg + 1)
    x0, x1 = S.gens()
    return 1 / (1 - x0 - x1 + x0 * x1)

# Final conjectural graph sum expressing DR_g(1,-1)


def DR11_sum(g, deg=None, **kwds):
    if deg is None:
        deg = g
    return graph_sum(g, 2, decgraphs=DR11_graphs(g, maxdeg=deg), globalfact=divbyaut, vertterm=DR11_vterm, edgeterm=DR11_eterm, deg=deg, **kwds)


def DR11_decgraphs(g, maxdeg=None):
    """
    Decorate the graphs of DR11_graphs(g) with a choice of half-edge
    at each non-root vertex.
    """
    for gr0, _ in DR11_graphs(g, maxdeg):
        L = [v for v in gr0.legs if 1 not in v]
        for choice in itertools.product(*L):
            yield (gr0, choice)


def divbyaut_new(gamma, dec):
    zeroverts = gamma.genera.count(0)
    return factorial(zeroverts - 1) / gamma.automorphism_number()

# Vertex- and edgeterms for DR11


def DR11_vterm_new(gv, nv, maxdeg, **kwargs):
    R = TautologicalRing(gv, nv)
    if gv == 0:
        gamma = kwargs['gamma']
        v = kwargs['v']
        f = R.fundamental_class()
        if 1 not in gamma.legs[v]:
            # we are in genus zero vertex not equal to base
            return -1 * f
        else:
            # we are at the base vertex
            return f
    else:
        return sum([(-1)**j * R.lambdaclass(j) for j in range(maxdeg + 1)])


def DR11_eterm_new(maxdeg, **kwargs):  # smarter: edterm(maxdeg=None,gamma=None,dec=None,e=None, *args, *kwds)
    S = PowerSeriesRing(QQ, 'x0,x1', default_prec=maxdeg + 1)
    x0, x1 = S.gens()
    e = kwargs['e']
    dec = kwargs['dec']
    ex = 0
    for i in e:
        if i in dec:
            ex += 1
    return 1 / ((1 - x0 - x1)**ex)

# Final conjectural graph sum expressing DR_g(1,-1)


def DR11_sum_new(g, deg=None, **kwds):
    if deg is None:
        deg = g
    return graph_sum(g, 2, decgraphs=DR11_decgraphs(g, maxdeg=deg), globalfact=divbyaut_new, vertterm=DR11_vterm_new, edgeterm=DR11_eterm_new, deg=deg, **kwds)


###
# Example 2 : Chiodo's formula from [JPPZ, Corollary 4]
###
def dicunion(d1, d2):
    r"""
    Computes the union of dictionaries d1, d2.

    EXAMPLES::

        sage: from admcycles.graph_sum import dicunion
        sage: d1 = {1:2, 3:4}; d2 = {1:2, 4:5};
        sage: dicunion(d1, d2)
        {1: 2, 3: 4, 4: 5}
    """
    d3 = copy(d1)
    d3.update(d2)
    return d3


def GammaWlist(g, Avector, k, r):
    r"""
    Returns a generator of pairs (Gamma, w) of pairs of stable graphs Gamma in genus g
    and admissible k-weightings w mod r on Gamma for the weight vector Avector.

    EXAMPLES::

        sage: from admcycles.graph_sum import GammaWlist
        sage: list(GammaWlist(0, (4,-1,-2,-3), 1, 2))
        [([0] [[1, 2, 3, 4]] [], {1: 4, 2: -1, 3: -2, 4: -3}),
         ([0, 0] [[1, 2, 5], [3, 4, 6]] [(5, 6)],
          {1: 4, 2: -1, 3: -2, 4: -3, 5: 0, 6: 0}),
         ([0, 0] [[1, 3, 5], [2, 4, 6]] [(5, 6)],
          {1: 4, 2: -1, 3: -2, 4: -3, 5: 1, 6: 1}),
         ([0, 0] [[1, 4, 5], [2, 3, 6]] [(5, 6)],
          {1: 4, 2: -1, 3: -2, 4: -3, 5: 0, 6: 0})]
    """
    n = len(Avector)
    stdic = {i + 1: Avector[i] for i in range(n)}
    for e in range(3 * g - 3 + n + 1):  # number of edges"""
        for gamma in list_strata(g, n, e):
            edges = gamma.edges()
            for w in itertools.product(*[list(range(r)) for h in edges]):
                dic = dicunion(stdic, {edges[i][0]: w[i] for i in range(len(edges))})
                dic.update({edges[i][1]: (r - w[i]) % r for i in range(len(edges))})
                if all((k * (2 * gv - 2 + len(hev)) - sum(dic[h] for h in hev)) % r == 0 for gv, hev in zip(gamma.genera(), gamma.legs())):
                    yield (gamma, dic)


def Chiodo_GF(g, r):
    def GF(gamma, dic):
        h1 = gamma.num_edges() - gamma.num_verts() + 1
        return r**(2 * g - 1 - h1) / gamma.automorphism_number()
    return GF


def expclass(x, g=None, n=None):
    r"""
    Given a tautological class x on Mbar_{g,n} of degree at least 1, returns
    the exponential exp(x) = 1 + x + 1/2*x^2 + ... of x.

    EXAMPLES::

        sage: from admcycles.graph_sum import expclass
        sage: from admcycles import TautologicalRing
        sage: R = TautologicalRing(1, 2)
        sage: expclass(R.psi(1))
        doctest:...: DeprecationWarning: expclass is deprecated. Please use the exp method of TautologicalClass instead
        See https://gitlab.com/modulispaces/admcycles/-/merge_requests/109 for details.
        Graph :      [1] [[1, 2]] []
        Polynomial : 1 + psi_1 + 1/2*psi_1^2

    TESTS::

        sage: from admcycles import TautologicalRing
        sage: R = TautologicalRing(0, 3)
        sage: expclass(R.zero(), 0, 3)
        Graph :      [0] [[1, 2, 3]] []
        Polynomial : 1
    """
    from .superseded import deprecation
    deprecation(109, 'expclass is deprecated. Please use the exp method of TautologicalClass instead')
    if g is not None:
        if n is not None:
            R = TautologicalRing(g, n)
        else:
            R = TautologicalRing(g)
        return R(x).exp()
    return x.exp()


def Chiodo_vterm(k, r):
    def vterm(gv, nv, maxdeg, **kwargs):
        R = TautologicalRing(gv, nv)
        expo = -sum((-1)**(m - 1) * bernoulli_polynomial(k / r, m + 1) / m / (m + 1) * R.kappa(m)
                    for m in range(1, 3 * gv - 3 + nv + 1))
        return R(expo).exp()
    return vterm


def Chiodo_legterm(r):
    def legterm(gv, nv, i, maxdeg, **kwargs):
        R = TautologicalRing(gv, nv)
        dec = kwargs['dec']
        origleg = kwargs['origleg']
        ai = dec[origleg]
        expo = sum((-1)**(m - 1) * bernoulli_polynomial(ai / r, m + 1) / m /
                   (m + 1) * R.psi(i)**m for m in range(1, 3 * gv - 3 + nv + 1))
        return R(expo).exp()
    return legterm


def Chiodo_edgeterm(r):
    def eterm(maxdeg, **kwargs):
        dec = kwargs['dec']
        e = kwargs['e']
        wh = dec[e[0]]

        S = PowerSeriesRing(QQ, 'x0,x1', default_prec=maxdeg + 4)
        x0, x1 = S.gens()
        expo = sum((-1)**(m - 1) * bernoulli_polynomial(wh / r, m + 1) / m /
                   (m + 1) * (x0**m - (-x1)**m) for m in range(1, maxdeg + 3))
        return (1 - exp(expo)) / (x0 + x1)
    return eterm


def Chiodo_alt(g, Avector, k, r):
    r"""
    Computes the mixed-degree class epsilon_* c(-R^* pi_* L) from [JPPZ17]_, Corollary 4.

    This agrees with the corresponding sum of DR_cycles with chiodo_coeff=True, weighted by appropriate powers of r.

    EXAMPLES::

        sage: from admcycles.graph_sum import Chiodo, Chiodo_alt
        sage: g=1; A=(0,0); k=2; r=2;
        sage: (Chiodo_alt(g, A, k, r)-Chiodo(g, r, k, A, 1)).simplify()
        0
        sage: g=1; A=(1,1); k=4; r=3;
        sage: (Chiodo_alt(g, A, k, r)-Chiodo(g, r, k, A, 1)).simplify()
        0
    """
    n = len(Avector)
    GWlist = GammaWlist(g, Avector, k, r)
    GF = Chiodo_GF(g, r)
    vterm = Chiodo_vterm(k, r)
    lterm = Chiodo_legterm(r)
    eterm = Chiodo_edgeterm(r)

    return graph_sum(g, n, decgraphs=GWlist, globalfact=GF, vertterm=vterm, legterm=lterm, edgeterm=eterm)


def Chiodo(g, r, k, A, x):
    n = len(A)
    return sum((r**(2 * g - 2 * d - 1)) * (x**d) * DR_cycle(g, A, d, k, chiodo_coeff=True, r_coeff=r) for d in range(3 * g - 3 + n + 1))



class kSignature:
    """
    A signature of a stratum of k-differentials.

    Attributes:
        sig (tuple): signature tuple
        g (int): genus
        n (int): total number of points
        p (int): number of poles
        z (int): number of zeroes
        poles (tuple): tuple of pole orders
        zeroes (tuple): tuple of zero orders
        pole_ind (tuple): tuple of indices of poles
        zero_ind (tuple): tuple of indices of zeroes

    EXAMPLES::

        sage: from admcycles.diffstrata.sig import kSignature
        sage: sig=kSignature((2,1,-1,0), k=1)
        sage: sig.g
        2
        sage: sig.n
        4
        sage: sig.poles
        (-1,)
        sage: sig.zeroes
        (2, 1)
        sage: sig.pole_ind
        (2,)
        sage: sig.zero_ind
        (0, 1)
        sage: sig.p
        1
        sage: sig.z
        2
    """

    def __init__(self, sig, k):
        """
        Initialise signature

        Args:
            sig (tuple): signature tuple of integers adding up to k*(2g-2)
            k (int): order of the differential.
        """
        self.sig = tuple(sig)
        sum_sig = sum(sig)
        if sum_sig % (2 * k) != 0:
            raise ValueError("Error! Illegal signature: Genus not an integer")
        self.g = int(sum_sig / (2 * k)) + 1
        self.n = len(sig)
        self.k = k
        self.poles = tuple(p for p in sig if p < 0)
        self.zeroes = tuple(z for z in sig if z > 0)
        self.marked_points = tuple(k for k in sig if k == 0)
        self.p = len(self.poles)
        self.z = len(self.zeroes)
        self.pole_ind = tuple(i for i, p in enumerate(sig) if p < 0)
        self.zero_ind = tuple(i for i, z in enumerate(sig) if z > 0)

    def __repr__(self):
        if self.k == 1:
            return "Signature(%r)" % (self.sig,)
        else:
            return "Signature(%r, k=%r)" % (self.sig, self.k)

    def __hash__(self):
        return hash((self.sig, self.k))

    def __eq__(self, other):
        try:
            return self.sig == other.sig
        except AttributeError:
            return False


class Signature(kSignature):
    """
    A signature of an abelian stratum.

    Attributes:
        sig (tuple): signature tuple
        g (int): genus
        n (int): total number of points
        p (int): number of poles
        z (int): number of zeroes
        poles (tuple): tuple of pole orders
        zeroes (tuple): tuple of zero orders
        pole_ind (tuple): tuple of indices of poles
        zero_ind (tuple): tuple of indices of zeroes

    EXAMPLES::

        sage: from admcycles.diffstrata.sig import Signature
        sage: sig=Signature((2,1,-1,0))
        sage: sig.g
        2
        sage: sig.n
        4
        sage: sig.poles
        (-1,)
        sage: sig.zeroes
        (2, 1)
        sage: sig.pole_ind
        (2,)
        sage: sig.zero_ind
        (0, 1)
        sage: sig.p
        1
        sage: sig.z
        2
    """

    def __init__(self, sig):
        super().__init__(sig, k=1)


class QuadraticSignature(kSignature):
    """
    A signature of an quadratic stratum.

    Attributes:
        sig (tuple): signature tuple
        difftype (string): either "p" for primitive or "gs" of global square
        g (int): genus
        n (int): total number of points
        p (int): number of poles
        z (int): number of zeroes
        poles (tuple): tuple of pole orders
        zeroes (tuple): tuple of zero orders
        pole_ind (tuple): tuple of indices of poles
        zero_ind (tuple): tuple of indices of zeroes

    EXAMPLES::

        sage: from admcycles.diffstrata.sig import QuadraticSignature
        sage: sig=QuadraticSignature((4,1,-1,0), "p")
        sage: sig.difftype
        'p'
        sage: sig.g
        2
        sage: sig.n
        4
        sage: sig.poles
        (-1,)
        sage: sig.zeroes
        (4, 1)
        sage: sig.pole_ind
        (2,)
        sage: sig.zero_ind
        (0, 1)
        sage: sig.p
        1
        sage: sig.z
        2
    """

    def __init__(self, sig, difftype):
        super().__init__(sig, k=2)
        assert difftype in ["p", "gs"]
        if difftype == "gs" and any(a % 2 != 0 for a in sig):
            raise ValueError(
                "Quadratic differentials with odd-order points can't be global squares.")
        self.difftype = difftype

    def __repr__(self):
        return "QuadraticSignature(%r, %r)" % (self.sig, self.difftype)

    def __hash__(self):
        return hash((self.sig, self.k, self.difftype))



r"""
Mixcellaneous functions
"""
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.combinat.subset import Subsets
from sage.combinat.combination import Combinations


ENABLE_DPRINT = False
ENABLE_DSAVE = False

A_list = [ZZ(6 * n).factorial() / (ZZ(3 * n).factorial() * ZZ(2 * n).factorial())
          for n in range(100)]
B_list = [ZZ(6 * n + 1).factorial() / ((6 * n - 1) * ZZ(3 * n).factorial() * ZZ(2 * n).factorial())
          for n in range(100)]


def get_memory_usage():
    """
    Return the memory usage of the current process in megabytes.

    This function was part of sage.misc.getusage but the module was
    removed in sage 9.5

    OUTPUT: a float representing the number of megabytes used.

    EXAMPLES::

        sage: from admcycles.DR.utils import get_memory_usage
        sage: t = get_memory_usage(); t  # random
        873.98046875
        sage: type(t)
        <... 'float'>
    """
    import psutil
    return psutil.Process().memory_info().vms / float(1048576)


def dprint(string, *args):
    if ENABLE_DPRINT:
        print(string % args)


def dsave(string, *args):
    if ENABLE_DSAVE:
        from sage.misc.persist import save
        save(0, string % args)


def aut(L):
    """
    Return the cardinality of the automorphism group of the list ``L``.

    EXAMPLES::

       sage: from admcycles.DR.utils import aut
       sage: aut([])
       1
       sage: aut([4,1,3,2])
       1
       sage: aut([4,5,6,5,4,4,6])
       24
    """
    if not L:
        return ZZ.one()
    L.sort()
    total = ZZ.one()
    n = 1
    last = L[0]
    for l in L[1:]:
        if l == last:
            n += 1
            total *= n
        else:
            n = 1
            last = l
    return total


def remove_duplicates(L):
    """
    Remove duplicate elements in a list ``L``.

    One cannot use ``set(L)`` because the elements of ``L`` are not hashable.

    INPUT:

    - ``L`` -- a list

    OUTPUT:

    a list

    EXAMPLES::

       sage: from admcycles.DR.utils import remove_duplicates
       sage: remove_duplicates([4,7,6,4,3,3,4,2,2,1])
       [1, 2, 3, 4, 6, 7]
    """
    if not L:
        return L
    L.sort()
    LL = [L[0]]
    for i, Li in enumerate(L[1:]):
        if Li != L[i]:
            LL.append(Li)
    return LL


def subsequences(n, l, symm):
    """
    Return all subsequences of length ``l`` of ``n`` points with symmetry in the first ``symm`` points.

    EXAMPLES::

        sage: from admcycles.DR.utils import subsequences
        sage: subsequences(5,2,2)
        [[2, 3], [2, 4], [3, 4], [1, 2], [1, 3], [1, 4], [1, 1]]
    """
    sym = max(symm, 1)
    answer = []
    for ones in range(min(l, sym) + 1):
        for others in Subsets(tuple(range(2, n - sym + 2)), l - ones):
            answer.append([1 for _ in range(ones)] + sorted(list(others)))
    return answer


def interpolate(A, B, var='x'):
    r"""
    Univariate Lagrange interpolation over the rationals.

    EXAMPLES::

        sage: from admcycles.DR.utils import interpolate
        sage: p = interpolate([1/2, -2, 3], [4/5, 2/3, -7/6])
        sage: p(1/2)
        4/5
        sage: p(-2)
        2/3
        sage: p(3)
        -7/6

    TESTS::

        sage: from admcycles.DR.utils import interpolate
        sage: parent(interpolate([], []))
        Univariate Polynomial Ring in x over Rational Field
        sage: parent(interpolate([], [], 'r'))
        Univariate Polynomial Ring in r over Rational Field
    """
    if len(A) != len(B):
        raise ValueError
    return QQ[var].lagrange_polynomial(zip(A, B))


def simplify_sparse(vec):
    """
    Collect coefficients in a list of pairs (index, coefficient).

    This also sorts the indices and removes indices with zero coefficient.

    EXAMPLES::

        sage: from admcycles.DR.utils import simplify_sparse
        sage: simplify_sparse([('b',6),('a',1),('c',2),('a',-1),('b',5)])
        [['b', 11], ['c', 2]]
    """
    vec.sort()
    vec2 = []
    last_index = None
    for index, coeff in vec:
        if index == last_index:
            if vec2[-1][1] == -coeff:
                vec2.pop()
                last_index = None
            else:
                vec2[-1][1] += coeff
        else:
            vec2.append([index, coeff])
            last_index = index
    return vec2


def setparts_recur(symlist, progress):
    if not symlist:
        yield progress
        return
    for i in Combinations(symlist[1:]):
        j = [symlist[0]] + i
        if progress and j < progress[-1]:
            continue
        cur = 0
        new_symlist = []
        for k in range(len(symlist)):
            if cur < len(j) and symlist[k] == j[cur]:
                cur += 1
            else:
                new_symlist.append(symlist[k])
        yield from setparts_recur(new_symlist, progress + [j])


def setparts_with_auts(symlist):
    r"""
    Iterate through the pairs ``(part, aut)`` where ``part`` is a set
    partition of ``symlist`` and ``aut`` is its number of
    automorphisms.

    EXAMPLES::

        sage: from admcycles.DR.utils import setparts_with_auts

        sage: list(setparts_with_auts(symlist=[1]))
        [([[1]], 1)]
        sage: list(setparts_with_auts(symlist=[2]))
        [([[2]], 1)]
        sage: list(setparts_with_auts(symlist=[1, 1]))
        [([[1], [1]], 1), ([[1, 1]], 1)]
        sage: list(setparts_with_auts(symlist=[3]))
        [([[3]], 1)]
        sage: list(setparts_with_auts(symlist=[1, 2]))
        [([[1], [2]], 1), ([[1, 2]], 1)]
        sage: list(setparts_with_auts(symlist=[1, 1, 1]))
        [([[1], [1], [1]], 1), ([[1], [1, 1]], 3), ([[1, 1, 1]], 1)]
    """
    a = aut(symlist)
    for i in setparts_recur(symlist, []):
        b = aut(i)
        for j in i:
            b *= aut(j)
        yield (i, a // b)



# -*- coding: utf-8 -*-
# ****************************************************************************
#       Copyright (C) 2021, Vincent Delecroix <20100.delecroix@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

def inverse_of_unit(elt):
    """
    Generic inversion of univariate or multivariate element of a graded algebra.

    TESTS::

        sage: Integers(1)['x'](0).inverse_of_unit()
        0
    """
    P = elt.parent()
    if not elt.is_unit():
        raise ArithmeticError("{} is not a unit in {}".format(elt, elt.parent()))

    u = elt.constant_coefficient()
    ui = u.inverse_of_unit()
    n = - ui * (elt - u)  # nilpotent
    nn = n
    res = P.one()
    while nn:
        res += nn
        nn *= n
    return ui * res



r"""
Multiplication in the strata algebra.
"""

import itertools

from sage.misc.cachefunc import cached_function
from sage.rings.integer_ring import ZZ
from sage.modules.free_module_element import vector


from .graph import Graph, R, X, all_strata, all_pure_strata, num_strata, single_stratum, unpurify_map, contraction_table, pure_strata_autom_count, automorphism_cosets, graph_isomorphic, num_of_stratum
from .moduli import MODULI_SM, MODULI_CT, MODULI_ST, MODULI_SMALL, dim_form
from .utils import simplify_sparse, setparts_with_auts


def get_marks(n, symm):
    if symm == 0:
        return tuple(range(1, n + 1))
    return (1,) * symm + tuple(range(2, n - symm + 2))


@cached_function
def multiply(r1, i1, r2, i2, g, rmax, markings=(), moduli_type=MODULI_ST):
    r"""
    Return the result of the multiplication of the ``(r1, i1)``-th generator
    with the ``(r2, i2)`` one.

    The output is a list representing the coefficients with respect to the
    generators in rank ``r1+r2``.

    EXAMPLES::

        sage: from admcycles.DR.graph import num_strata
        sage: from admcycles.DR.algebra import multiply
        sage: r1 = 1
        sage: r2 = 2
        sage: for i1 in range(num_strata(2, r1)):
        ....:     for i2 in range(num_strata(2, r2)):
        ....:         print(i1, i2, multiply(r1, i1, r2, i2, 2, r1 + r2))
        0 0 [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        0 1 [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        0 2 [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        0 3 [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        0 4 [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        0 5 [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0]
        0 6 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0]
        0 7 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0]
        1 0 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        1 1 [0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        1 2 [0, 0, 0, 0, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        1 3 [0, 0, 0, 0, 0, -2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        1 4 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0]
        1 5 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        1 6 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -2, 0, 0, 0, 0]
        1 7 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0]
        2 0 [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        2 1 [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]
        2 2 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0]
        2 3 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0]
        2 4 [0, 0, 0, 0, 0, 0, 0, 0, -4, 0, 0, 0, 0, 1, 0, 0, 0]
        2 5 [0, 0, 0, 0, 0, 0, 0, 0, 0, -2, -2, 0, 0, 0, 1, 0, 0]
        2 6 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0]
        2 7 [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -8, 0, 4]
    """
    unpurify = unpurify_map(g, r1 + r2, markings, moduli_type)
    gens = all_strata(g, r1 + r2, markings, moduli_type)
    ngens = num_strata(g, r1 + r2, markings, moduli_type)
    answer = [0 for i in range(ngens)]
    pure_strata = [all_pure_strata(g, r, markings, moduli_type)
                   for r in range(rmax + 1)]
    contraction_dict = contraction_table(g, rmax, markings, moduli_type)
    G1 = single_stratum(i1, g, r1, markings, moduli_type)
    G2 = single_stratum(i2, g, r2, markings, moduli_type)
    G1copy = Graph(G1.M)
    G2copy = Graph(G2.M)
    G1copy.purify()
    G2copy.purify()
    pure_r1 = G1copy.num_edges() - len(markings)
    pure_r2 = G2copy.num_edges() - len(markings)
    found = False
    for i in range(len(pure_strata[pure_r1])):
        if G1copy.M == pure_strata[pure_r1][i].M:
            G1_key = (pure_r1, i)
            found = True
            break
    assert found, "G1 purification failed"
    found = False
    for i in range(len(pure_strata[pure_r2])):
        if G2copy.M == pure_strata[pure_r2][i].M:
            G2_key = (pure_r2, i)
            found = True
            break
    assert found, "G2 purification failed"
    if G1_key > G2_key:
        return multiply(r2, i2, r1, i1, g, rmax, markings, moduli_type)

    if (G1_key, G2_key) not in contraction_dict:
        return answer
    for L in contraction_dict[(G1_key, G2_key)]:
        H = pure_strata[L[0][0]][L[0][1]]
        Hloops = []
        if moduli_type > MODULI_CT:
            for i in range(1, H.M.nrows()):
                for j in range(1, H.M.ncols()):
                    if H.M[i, j][0] == 2:
                        Hloops.append((i, j))
        auts = pure_strata_autom_count(
            L[0][1], g, L[0][0], markings, moduli_type)
        B = L[1]
        if len(B) == pure_r1 and len(B) == pure_r2:
            auts *= 2
        aut_cosets1 = automorphism_cosets(i1, g, r1, markings, moduli_type)
        aut_cosets2 = automorphism_cosets(i2, g, r2, markings, moduli_type)
        auts /= aut_cosets1[0] * aut_cosets2[0]
        for isom1 in aut_cosets1[1]:
            for isom2 in aut_cosets2[1]:
                Hcopy = Graph(H.M)
                vmap1 = [0 for i in range(G1.M.nrows())]
                for i in range(1, G1.M.nrows()):
                    vmap1[i] = L[2][0][L[4][0][isom1[0]
                                               [i - 1] - 1] - 1]
                emap1 = [0 for i in range(G1.M.ncols())]
                for i in range(1, G1.M.ncols()):
                    emap1[i] = L[2][1][L[4][1][isom1[1]
                                               [i - 1] - 1] - 1]
                vmap2 = [0 for i in range(G2.M.nrows())]
                for i in range(1, G2.M.nrows()):
                    vmap2[i] = L[3][0][L[5][0][isom2[0]
                                               [i - 1] - 1] - 1]
                emap2 = [0 for i in range(G2.M.ncols())]
                for i in range(1, G2.M.ncols()):
                    emap2[i] = L[3][1][L[5][1][isom2[1]
                                               [i - 1] - 1] - 1]

                psilooplist = []
                psiindexlist = []
                loop_factor = ZZ.one()
                for i in range(1, G1.M.nrows()):
                    for j in range(1, G1.M.ncols()):
                        if G1.M[i, j][0] != 0:
                            if G1.M[i, j][0] == 1:
                                if G1.M[i, j][1] != 0:
                                    jj = emap1[j]
                                    for ii in vmap1[i]:
                                        if H.M[ii, jj] != 0:
                                            Hcopy.M[ii, jj] += G1.M[i, j][1] * X
                                            break
                            elif G1.M[i, j][1] == 0:
                                loop_factor *= 2
                            else:
                                jj = emap1[j]
                                psilooplist.append([[G1.M[i, j][1], G1.M[i, j][2]], [
                                                   G1.M[i, j][2], G1.M[i, j][1]]])
                                psiindexlist.append([jj])
                                for ii in vmap1[i]:
                                    for k in range(H.M[ii, jj][0]):
                                        psiindexlist[-1].append(ii)
                for i in range(1, G2.M.nrows()):
                    for j in range(1, G2.M.ncols()):
                        if G2.M[i, j][0] != 0:
                            if G2.M[i, j][0] == 1:
                                if G2.M[i, j][1] != 0:
                                    if G2.M[i, j][0] == 1:
                                        jj = emap2[j]
                                        for ii in vmap2[i]:
                                            if H.M[ii, jj] != 0:
                                                Hcopy.M[ii,
                                                        jj] += G2.M[i, j][1] * X
                                                break
                            elif G2.M[i, j][1] == 0:
                                loop_factor *= 2
                            else:
                                jj = emap2[j]
                                psilooplist.append([[G2.M[i, j][1], G2.M[i, j][2]], [
                                                   G2.M[i, j][2], G2.M[i, j][1]]])
                                psiindexlist.append([jj])
                                for ii in vmap2[i]:
                                    for k in range(H.M[ii, jj][0]):
                                        psiindexlist[-1].append(ii)

                Klocationlist = []
                Kindexlist = []
                for i in range(1, G1.M.nrows()):
                    for r in range(1, rmax + 1):
                        for k in range(G1.M[i, 0][r]):
                            Klocationlist.append(vmap1[i])
                            Kindexlist.append(r)
                for i in range(1, G2.M.nrows()):
                    for r in range(1, rmax + 1):
                        for k in range(G2.M[i, 0][r]):
                            Klocationlist.append(vmap2[i])
                            Kindexlist.append(r)

                psilist = []
                for j in B:
                    S = [i for i in range(1, H.M.nrows()) if H.M[i, j][0] != 0]
                    if len(S) == 2:
                        psilist.append([[S[0], j], [S[1], j]])
                    else:
                        psilooplist.append([[0, 1], [1, 0]])
                        psiindexlist.append([j, S[0], S[0]])

                for psiloopvals in itertools.product(*psilooplist):
                    for Klocs in itertools.product(*Klocationlist):
                        for psilocs in itertools.product(*psilist):
                            Hcopycopy = Graph(Hcopy.M)
                            for i in range(len(psiindexlist)):
                                Hcopycopy.M[psiindexlist[i][1],
                                            psiindexlist[i][0]] += psiloopvals[i][0] * X
                                if psiindexlist[i][1] == psiindexlist[i][2]:
                                    Hcopycopy.M[psiindexlist[i][1],
                                                psiindexlist[i][0]] += psiloopvals[i][1] * X**2
                                else:
                                    Hcopycopy.M[psiindexlist[i][2],
                                                psiindexlist[i][0]] += psiloopvals[i][1] * X
                            for i in range(len(Kindexlist)):
                                Hcopycopy.M[Klocs[i],
                                            0] += X**Kindexlist[i]
                            for i in psilocs:
                                Hcopycopy.M[i[0], i[1]] += X
                            for k in Hloops:
                                if Hcopycopy.M[k][2] > Hcopycopy.M[k][1]:
                                    Hcopycopy.M[k] += (Hcopycopy.M[k]
                                                       [2] - Hcopycopy.M[k][1]) * (X - X**2)
                            Hcopycopy.compute_invariant()
                            for which_gen in unpurify[L[0]]:
                                if graph_isomorphic(Hcopycopy, gens[which_gen]):
                                    answer[which_gen] += (-1)**len(B) * \
                                        loop_factor / auts
                                    break
    return answer


def check_associativity(g, r1, r2, r3, markings=(), moduli_type=MODULI_ST):
    r"""
    Check associativity of the strata algebra.

    If it fails a ``ValueError`` is raised.

    EXAMPLES::

        sage: from admcycles.DR.algebra import check_associativity
        sage: check_associativity(2, 1, 1, 1)
    """
    ngens1 = num_strata(g, r1, markings, moduli_type)
    ngens2 = num_strata(g, r2, markings, moduli_type)
    ngens3 = num_strata(g, r3, markings, moduli_type)
    for i1, i2, i3 in itertools.product(range(ngens1), range(ngens2), range(ngens3)):
        a = multiply(r1, i1, r2, i2, g, r1 + r2 +
                     r3, markings, moduli_type)
        answer1 = vector(
            [0 for i in range(num_strata(g, r1 + r2 + r3, markings, moduli_type))])
        for j in range(num_strata(g, r1 + r2, markings, moduli_type)):
            if a[j] == 0:
                continue
            answer1 += a[j] * vector(multiply(r1 + r2, j, r3,
                                              i3, g, r1 + r2 + r3, markings, moduli_type))
        a = multiply(r1, i1, r3, i3, g, r1 + r2 +
                     r3, markings, moduli_type)
        answer2 = vector(
            [0 for i in range(num_strata(g, r1 + r2 + r3, markings, moduli_type))])
        for j in range(num_strata(g, r1 + r3, markings, moduli_type)):
            if a[j] == 0:
                continue
            answer2 += a[j] * vector(multiply(r1 + r3, j, r2,
                                              i2, g, r1 + r2 + r3, markings, moduli_type))
        if answer1 != answer2:
            raise ValueError("i1=%s i2=%s i3=%s" % (i1, i2, i3))


@cached_function
def kappa_conversion(sigma):
    r"""
    EXAMPLES::

        sage: from admcycles.DR.algebra import kappa_conversion
        sage: kappa_conversion((1, 1, 1))
        [[3*X, 1], [X^2 + X, 3], [X^3, 2]]
    """
    answer = []
    for spart in setparts_with_auts(list(sigma)):
        coeff = spart[1]
        poly = R(0)
        for part in spart[0]:
            coeff *= ZZ(len(part) - 1).factorial()
            poly += X**sum(part)
        answer.append([poly, coeff])
    return answer


@cached_function
def kappa_conversion_inverse(sigma):
    r"""
    EXAMPLES::

        sage: from admcycles.DR.algebra import kappa_conversion_inverse
        sage: kappa_conversion_inverse((1, 1, 1))
        [[3*X, 1], [X^2 + X, -3], [X^3, 1]]
    """
    answer = []
    for spart in setparts_with_auts(list(sigma)):
        coeff = spart[1] * (-1)**(len(sigma) - len(spart[0]))
        poly = R(0)
        for part in spart[0]:
            poly += X**sum(part)
        answer.append([poly, coeff])
    return answer


@cached_function
def convert_to_monomial_basis(num, g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    EXAMPLES::

        sage: from admcycles.DR.algebra import convert_to_monomial_basis

        sage: convert_to_monomial_basis(1, 2, 3, (1, 2, 3))
        [(1, 1), (0, 1)]
        sage: convert_to_monomial_basis(2, 2, 3, (1, 2, 3))
        [(2, 1), (1, 3), (0, 2)]
    """
    answer = []
    G = single_stratum(num, g, r, markings, moduli_type)
    genus_vec = []
    kappa_vec = []
    for i in range(1, G.M.nrows()):
        genus_vec.append(G.M[i, 0][0])
        kappa_vec.append([])
        for j in range(1, r + 1):
            for k in range(G.M[i, 0][j]):
                kappa_vec[-1].append(j)
        kappa_vec[-1] = kappa_conversion(tuple(kappa_vec[-1]))
    for choice in itertools.product(*kappa_vec):
        coeff = 1
        GG = Graph(G.M)
        for i in range(1, G.M.nrows()):
            GG.M[i, 0] = genus_vec[i - 1] + choice[i - 1][0]
            coeff *= choice[i - 1][1]
        answer.append((num_of_stratum(GG, g, r, markings, moduli_type), coeff))
    return answer


# NOTE: this function is not used anywhere
@cached_function
def convert_to_pushforward_basis(num, g, r, markings=(), moduli_type=MODULI_ST):
    answer = []
    G = single_stratum(num, g, r, markings, moduli_type)
    genus_vec = []
    kappa_vec = []
    for i in range(1, G.M.nrows()):
        genus_vec.append(G.M[i, 0][0])
        kappa_vec.append([])
        for j in range(1, r + 1):
            for k in range(G.M[i, 0][j]):
                kappa_vec[-1].append(j)
        kappa_vec[-1] = kappa_conversion_inverse(
            tuple(kappa_vec[-1]))
    for choice in itertools.product(*kappa_vec):
        coeff = 1
        GG = Graph(G.M)
        for i in range(1, G.M.nrows()):
            GG.M[i, 0] = genus_vec[i - 1] + choice[i - 1][0]
            coeff *= choice[i - 1][1]
        answer.append((num_of_stratum(GG, g, r, markings, moduli_type), coeff))
    return answer


def convert_vector_to_monomial_basis(vec, g, r, markings=(), moduli_type=MODULI_ST):
    r"""
    EXAMPLES::

        sage: from admcycles.DR.moduli import MODULI_SM, MODULI_RT, MODULI_CT, MODULI_ST
        sage: from admcycles.DR.algebra import convert_vector_to_monomial_basis
        sage: convert_vector_to_monomial_basis(vec=(-24504480, 1663200, -36000), g=8, r=3, markings=(), moduli_type=MODULI_SM)
        [-22913280, 1555200, -36000]
        sage: convert_vector_to_monomial_basis(vec=(0, 0, -15, 0, 15, 15, 0, 0, 0), g=2, r=2, markings=(1, 2), moduli_type=MODULI_RT)
        [0, 0, -15, 0, 15, 15, 0, 0, 0]
        sage: convert_vector_to_monomial_basis(vec=(71820, -7020, 9720, -41580, 9288, -3240, -18900, 756), g=2, r=2, markings=(1,), moduli_type=MODULI_CT)
        [64800, -7020, 9720, -41580, 9288, -3240, -18900, 756]
        sage: convert_vector_to_monomial_basis(vec=(81/2, -45/2, -45/2, -45/2, 9/2, 9/2, 9/2, -27/2, -9/4), g=1, r=1, markings=(1, 2, 3), moduli_type=MODULI_ST)
        [81/2, -45/2, -45/2, -45/2, 9/2, 9/2, 9/2, -27/2, -9/4]
    """
    l = len(vec)
    vec2 = [0 for i in range(l)]
    for i in range(l):
        if vec[i] != 0:
            for x in convert_to_monomial_basis(i, g, r, markings, moduli_type):
                vec2[x[0]] += x[1] * vec[i]
    return vec2


# NOTE: this function is not used anywhere
def convert_vector_to_pushforward_basis(vec, g, r, markings=(), moduli_type=MODULI_ST):
    l = len(vec)
    vec2 = [0 for i in range(l)]
    for i in range(l):
        if vec[i] != 0:
            for x in convert_to_pushforward_basis(i, g, r, markings, moduli_type):
                vec2[x[0]] += x[1] * vec[i]
    return vec2


def kappa_multiple(vec, which_kappa, g, r, n, symm, moduli_type=MODULI_ST):
    r"""
    Return the multiplication of the sparse vector ``vec`` by a kappa-class.
    """
    vec2 = []
    for num, coeff in vec:
        for num2, coeff2 in single_kappa_multiple(num, which_kappa, g, r, n, symm, moduli_type):
            vec2.append([num2, coeff * coeff2])
    return simplify_sparse(vec2)


# TODO: why using (n, symm) rather than a tuple of markings here?
def psi_multiple(vec, which_psi, g, r, n, symm, moduli_type=MODULI_ST):
    r"""
    Return the multiplication of the sparse vector ``vec`` by a psi-class.
    """
    vec2 = []
    for num, coeff in vec:
        for num2, coeff2 in single_psi_multiple(num, which_psi, g, r, n, symm, moduli_type):
            vec2.append([num2, coeff * coeff2])
    return simplify_sparse(vec2)


# TODO: why using (n, symm) rather than a tuple of markings here?
def insertion_pullback(vec, new_mark, g, r, n, symm, moduli_type=MODULI_ST, from_small=False):
    r"""
    For a relation ``vec`` we return its pullback along the forgetfulmap that forgets a point.
    """
    vec2 = []
    for x in vec:
        for y in single_insertion_pullback(x[0], new_mark, g, r, n, symm, moduli_type, from_small):
            vec2.append([y[0], x[1] * y[1]])
    return simplify_sparse(vec2)


# TODO: why using (n, symm) rather than a tuple of markings here?
@cached_function
def single_psi_multiple(num, which_psi, g, r, n, symm, moduli_type=MODULI_ST):
    r"""
    Takes the index of a stratum ``num`` and returns the stratum in degree ``r+1``
    (as a sparse vector of length 1) corresponding to its multiplication by a psi-class.

    EXAMPLES::

        sage: from admcycles.DR.moduli import MODULI_RT, MODULI_CT, MODULI_ST
        sage: from admcycles.DR.algebra import single_psi_multiple
        sage: single_psi_multiple(5, 1, 2, 2, 2, 0, MODULI_ST)
        [(11, 1)]
    """
    markings = get_marks(n, symm)
    G = single_stratum(num, g, r, markings, moduli_type)
    answer = []
    for j in range(1, G.M.ncols()):
        if G.M[0, j] == which_psi:
            good_j = j
            break
    for i in range(1, G.M.nrows()):
        if G.M[i, good_j] != 0:
            deg = 0
            dim_used = 0
            for j in range(1, r + 1):
                dim_used += j * G.M[i, 0][j]
            for j in range(1, G.M.ncols()):
                dim_used += G.M[i, j][1] + G.M[i, j][2]
                deg += G.M[i, j][0]
            if dim_used < dim_form(G.M[i, 0][0], deg, moduli_type):
                GG = Graph(G.M)
                GG.M[i, good_j] += X
                answer.append(
                    (num_of_stratum(GG, g, r + 1, markings, moduli_type), 1))
            break
    return answer


# TODO: why using (n, symm) rather than a tuple of markings here?
@cached_function
def single_kappa_multiple(num, which_kappa, g, r, n, symm, moduli_type=MODULI_ST):
    r"""
    Takes the index of a stratum ``num`` and returns the index of its multiplication by a kappa-class.
    This uses the multi-indexed pushforward kappa-classes, rather than monomials in kappa-classes.
    """
    markings = get_marks(n, symm)
    G = single_stratum(num, g, r, markings, moduli_type)
    answer = []
    for i in range(1, G.M.nrows()):
        deg = 0
        dim_used = 0
        for j in range(1, r + 1):
            dim_used += j * G.M[i, 0][j]
        for j in range(1, G.M.ncols()):
            dim_used += G.M[i, j][1] + G.M[i, j][2]
            deg += G.M[i, j][0]
        if dim_used + which_kappa <= dim_form(G.M[i, 0][0], deg, moduli_type):
            GG = Graph(G.M)
            GG.M[i, 0] += X**which_kappa
            answer.append((num_of_stratum(GG, g, r + which_kappa,
                                          markings, moduli_type), 1))
            for j in range(1, r + 1):
                if G.M[i, 0][j] > 0:
                    GG = Graph(G.M)
                    GG.M[i, 0] += X**(j + which_kappa)
                    GG.M[i, 0] -= X**j
                    answer.append((num_of_stratum(
                        GG, g, r + which_kappa, markings, moduli_type), -G.M[i, 0][j]))
    return answer


# NOTE: this function is not used anywhere
def single_kappa_psi_multiple(num, kappa_partition, psi_exps, g, r, n=0, moduli_type=MODULI_ST):
    markings = tuple(range(1, n + 1))
    G = single_stratum(num, g, r, markings, moduli_type)
    GG = Graph(G.M)
    for j in range(1, G.M.ncols()):
        if GG.M[0, j] != 0:
            for i in range(1, G.M.nrows()):
                if GG.M[i, j] != 0:
                    GG.M[i, j] += psi_exps[GG.M[0, j][0] - 1] * X
                    break
    rnew = r + sum(kappa_partition) + sum(psi_exps)
    answer = []
    kappa_options = [list(range(1, G.M.nrows()))
                     for i in range(len(kappa_partition))]
    for kappa_distrib in itertools.product(*kappa_options):
        GGG = Graph(GG.M)
        for i in range(len(kappa_partition)):
            GGG.M[kappa_distrib[i], 0] += X**(kappa_partition[i])
        is_bad = False
        for i in range(1, GGG.M.nrows()):
            deg = 0
            dim_used = 0
            for j in range(1, rnew + 1):
                dim_used += j * GGG.M[i, 0][j]
            for j in range(1, GGG.M.ncols()):
                dim_used += GGG.M[i, j][1] + GGG.M[i, j][2]
                deg += GGG.M[i, j][0]
            if dim_used > dim_form(GGG.M[i, 0][0], deg, moduli_type):
                is_bad = True
                break
        if is_bad:
            continue
        answer.append(
            (num_of_stratum(GGG, g, rnew, markings, moduli_type), 1))
    return answer


# TODO: why using (n, symm) rather than markings here?
@cached_function
def single_insertion_pullback(num, new_mark, g, r, n, symm, moduli_type=MODULI_ST, from_small=False):
    r"""
    Takes the index of a stratum ``num`` and returns the sparse vector
    representation of its pullback along the forgetfulmap that foregets a
    point.

    EXAMPLES::

        sage: from admcycles.DR.algebra import single_insertion_pullback

        sage: single_insertion_pullback(0, 1, 5, 2, 0, 0, 1, False)
        [(0, 1), (3, -1)]
        sage: single_insertion_pullback(1, 1, 5, 2, 0, 0, 1, False)
        [(1, 1), (2, -1), (2, -1)]
        sage: single_insertion_pullback(10, 1, 2, 2, 1, 0, 3, False)
        [(25, 1), (27, -1)]
    """
    markings = get_marks(n, symm)
    if new_mark == 1:
        new_markings = get_marks(n + 1, symm + 1)
    else:
        new_markings = get_marks(n + 1, symm)
    if from_small:
        G = single_stratum(num, g, r, markings, MODULI_SMALL)
    else:
        G = single_stratum(num, g, r, markings, moduli_type)
    answer = []
    for i in range(1, G.M.nrows()):
        GG = Graph(G.M)
        if not (new_mark == 1 and symm > 0):
            for j in range(1, G.M.ncols()):
                if GG.M[0, j][0] >= new_mark:
                    GG.M[0, j] += 1
        GG.add_edge(i, 0, new_mark)
        answer.append(
            (num_of_stratum(GG, g, r, new_markings, moduli_type), 1))
        for j in range(1, r + 1):
            for k in range(GG.M[i, 0][j]):
                GGG = Graph(GG.M)
                GGG.M[i, 0] -= X**j
                GGG.M[i, -1] += j * X
                answer.append(
                    (num_of_stratum(GGG, g, r, new_markings, moduli_type), -1))
        if moduli_type <= MODULI_SM:
            continue
        for j in range(1, G.M.ncols()):
            if G.M[i, j][0] == 1:
                if G.M[i, j][1] >= 1:
                    x = G.M[i, j][1]
                    GGG = Graph(GG.M)
                    row1 = [GG.M[i, k] for k in range(GG.M.ncols())]
                    row2 = [0 for k in range(GG.M.ncols())]
                    row1[j] = 0
                    row1[-1] = 0
                    row2[j] = 1
                    row2[-1] = 1
                    GGG.split_vertex(i, row1, row2)
                    GGG.M[-2, -
                          1] += (x - 1) * X
                    answer.append(
                        (num_of_stratum(GGG, g, r, new_markings, moduli_type), -1))
            if not from_small:
                if G.M[i, j][0] == 2:
                    if G.M[i, j][1] >= 1 or G.M[i, j][2] >= 1:
                        x = G.M[i, j][1]
                        y = G.M[i, j][2]
                        row1 = [GG.M[i, k] for k in range(GG.M.ncols())]
                        row2 = [0 for k in range(GG.M.ncols())]
                        row1[j] = 0
                        row1[-1] = 0
                        row2[j] = 1
                        row2[-1] = 1
                        if y >= 1:
                            row1[j] = 1 + x * X
                            GGG = Graph(GG.M)
                            GGG.split_vertex(i, row1, row2)
                            GGG.M[-2, -
                                  1] += (y - 1) * X
                            answer.append(
                                (num_of_stratum(GGG, g, r, new_markings, moduli_type), -1))
                        if x >= 1:
                            row1[j] = 1 + y * X
                            GGG = Graph(GG.M)
                            GGG.split_vertex(i, row1, row2)
                            GGG.M[-2, -
                                  1] += (x - 1) * X
                            answer.append(
                                (num_of_stratum(GGG, g, r, new_markings, moduli_type), -1))
    return answer



import signal
import sys
import resource

if len(sys.argv) != 8:
    print('invalid arguments')
    sys.exit(3)

g = int(sys.argv[1])
n = int(sys.argv[2])
r = int(sys.argv[3])
moduli = int(sys.argv[4])
maxmem = int(sys.argv[5])
maxtime = int(sys.argv[6])
verbosity = int(sys.argv[7])

if verbosity >= 2:
    print(sys.argv)
    sys.stdout.flush()

MAX_VIRTUAL_MEMORY = maxmem * 1024 * 1024  # maxmem MB
resource.setrlimit(resource.RLIMIT_AS, (MAX_VIRTUAL_MEMORY, resource.RLIM_INFINITY))

try:
    import sage.all
    from admcycles.DR import FZ_methods_sanity_check
except (MemoryError, ImportError) as e:
    if verbosity >= 2:
        print(e, file=sys.stderr)
    sys.exit(2)

class TimeoutException(Exception):
    pass

def handler(signum, frame):
    raise TimeoutException("timeout")

signal.signal(signal.SIGALRM, handler)
signal.alarm(maxtime)

try:
    result = FZ_methods_sanity_check(g, r, n, moduli)
except TimeoutException as e:
    # time limit
    if verbosity >= 2:
        print(e, file=sys.stderr)
    sys.exit(2)
except MemoryError:
    # (probably) memory limit
    if verbosity >= 2:
        print(e, file=sys.stderr)
    sys.exit(3)
except Exception:
    # something
    if verbosity >= 2:
        print(e, file=sys.stderr)
    sys.exit(4)

sys.exit(0 if result else 1)
